#!/bin/bash

shellfu import pretty
shellfu import termcolors

shellfu import jats

#
# Core JAT utilities
#


#
# Path to main logfile
#
_JAT_CORE__RDIR=

#
# Colors for `jat__show_*` functions
#
_JAT_CORE__SCOLOR_INFO=
_JAT_CORE__SCOLOR_NONE=

#
# Show timestamps in terminal output?
#
# Note that if __JAT__DETERMINISTIC is 'true', all stamps will
# just be constant string.
#
_JAT_CORE__STAMPS=${_JAT_CORE__STAMPS:-}

#
# Colored words for `jat__show_*` functions
#
_JAT_CORE__SWORD_ASSERT=
_JAT_CORE__SWORD_END=
_JAT_CORE__SWORD_FAIL=
_JAT_CORE__SWORD_FINALIZE=
_JAT_CORE__SWORD_PASS=
_JAT_CORE__SWORD_PHASE=
_JAT_CORE__SWORD_RELOAD=
_JAT_CORE__SWORD_RUN=
_JAT_CORE__SWORD_START=
_JAT_CORE__SWORD_TERROR=
_JAT_CORE__SWORD_TEST=
_JAT_CORE__SWORD_TINFO=
_JAT_CORE__SWORD_TWARNING=

_jat_core__bumpid() {
    #
    # Create and/or bump ID named $1
    #
    # Print next ID from series named $1.  Series starts
    # with 1, so on firts call, value of ID is 1, then 2.
    # etc.
    #
    # If run is not initialized, ID value will be 0 and
    # return status will be 3.
    #
    local name=$1   # name of id
    local cache     # ordinal cache
    local ord       # ordinal
    if test -n "$_JAT_CORE__RDIR"; then
        cache=$(_jat_core__rd_path "series/$name")
    else
        echo 0; return 3
    fi
    if test -f "$cache"; then
        ord=$(<"$cache")
        ((ord++))
    else
        ord=1
    fi
    echo $ord >"$cache"
    echo "$name-$ord"
}

_jat_core__decorate_hint() {
    #
    # Decorate phase hint using phase type and case id from rundata
    #
    local phint
    local ptype
    local pcaseid
    phint=$(_jat_core__rd_keyr P.hint)
    ptype=$(_jat_core__rd_keyr P.type)
    pcaseid=$(_jat_core__rd_keyr P.pcaseid)
    test -n "$pcaseid" && echo -n "[$pcaseid] "
    echo -n "$phint"
    test -n "$phint" && echo -n " "
    echo "# $ptype"
}

_jat_core__rd_keya() {
    #
    # Append line $2 to run key $1
    #
    local key=$1    # key name
    local value=$2  # value to append
    local path      # real path
    path=$(_jat_core__rd_path "$key")
    echo "$value" >> "$path" || {
        _jat_core__show_error "error appending run data: key '$key' to '$path'"
        return 3
    }
}

_jat_core__rd_keyA() {
    #
    # Append line $2 to run key $1, unless already there
    #
    local key=$1    # key name
    local value=$2  # value to append
    local path      # real path
    path=$(_jat_core__rd_path "$key")
    grep -qsFxe "$value" "$path" && return 0     # already there
    echo "$value" >> "$path" || {
        _jat_core__show_error "error appending run data: key '$key' to '$path'"
        return 3
    }
}

_jat_core__rd_keyR() {
    #
    # Read run key $1 if it exists, return 1 otherwise
    #
    local key=$1    # key name
    local path      # real path
    path=$(_jat_core__rd_path "$key") || return 1
    cat "$path" || {
        _jat_core__show_error "error reading run data: key '$key' from '$path'"
        return 3
    }
}

_jat_core__rd_keyr() {
    #
    # Read run key $1
    #
    local key=$1    # key name
    local path      # real path
    path=$(_jat_core__rd_path "$key")
    cat "$path" || {
        _jat_core__show_error "error reading run data: key '$key' from '$path'"
        return 3
    }
}

_jat_core__rd_keyw() {
    #
    # Write $2 to run key $1
    #
    local key=$1    # key name
    local value=$2  # value to write
    local path      # real path
    path=$(_jat_core__rd_path "$key")
    echo "$value" > "$path" || {
        _jat_core__show_error "error writing run data: key '$key' to '$path'"
        return 3
    }
}

_jat_core__rd_path() {
    #
    # Dereference path to run key $1; true if exists
    #
    # Key starting with 'P.' will be phase-specific.
    #
    local key=$1    # key name
    local path      # real path
    path="$_JAT_CORE__RDIR/internal/"
    case $key in
        "")     _jat_core__show_error "no KEY?"; return 2 ;;
        P.*)    path+="phase-$(_jat_core__rd_keyr phase)/$key" ;;
        *)      path+="$key" ;;
    esac
    mkdir -p "${path%/*}" 2>/dev/null || {
        _jat_core__show_error "could not create key: $path"
        return 2
    }
    echo "$path"
    test -e "$path"
}

_jat_core__show() {
    #
    # Show messages $@ to user
    #
    local msg       # message to show (and keep)
    local stamp
    case $_JAT_CORE__STAMPS:$__JAT__DETERMINISTIC in
        false:*)    stamp="" ;;
        true:true)  stamp="0000-00-00T00:00:00+00:00" ;;
        true:false) stamp="$(date -Isec)" ;;
    esac
    test -n "$stamp" && stamp="[$_JAT_CORE__SCOLOR_INFO$stamp$_JAT_CORE__SCOLOR_NONE] "
    for msg in "$@"; do
        echo -e "$stamp$msg" >&2
    done
}

_jat_core__show_adetail() {
    #
    # Show assert detail
    #
    local msg=$1; shift
    local detail
    local sp=""     # indentation space, if needed
    test "$(_jat_core__rd_keyr phase)" == dummy || sp="  "
    _jat_core__show \
        "$sp$_JAT_CORE__SWORD_ASSERT.$_JAT_CORE__SWORD_TINFO: $msg"
    for detail in "$@"; do
        _jat_core__show \
            "$sp$_JAT_CORE__SWORD_ASSERT.$_JAT_CORE__SWORD_TINFO:  .. $detail"
    done
}

_jat_core__show_fail() {
    #
    # Show assert fail message $1 to stderr
    #
    _jat_core__show \
        "  $_JAT_CORE__SWORD_ASSERT.$_JAT_CORE__SWORD_FAIL: $1"
}

_jat_core__show_pass() {
    #
    # Show assert fail message $1 to stderr
    #
    _jat_core__show \
        "  $_JAT_CORE__SWORD_ASSERT.$_JAT_CORE__SWORD_PASS: $1"
}

_jat_core__show_pend() {
    #
    # Show phase end
    #
    local pdur=$1   # phase duration
    local pverd     # ^^  verdict text
    local phint     # ^^ hint
    phint=$(_jat_core__decorate_hint)
    case $(_jat_core__presult) in
        none)      pverd=""                            ;;
        pass)      pverd=".$_JAT_CORE__SWORD_PASS"     ;;
        fail)      pverd=".$_JAT_CORE__SWORD_FAIL"     ;;
        warning)   pverd=".$_JAT_CORE__SWORD_TWARNING" ;;
        error)     pverd=".$_JAT_CORE__SWORD_TERROR"   ;;
    esac
    _jat_core__show \
        "$_JAT_CORE__SWORD_PHASE.$_JAT_CORE__SWORD_END$pverd (d=$pdur) $phint" \
        ""
}

_jat_core__presult() {
    #
    # Print phase result
    #
    local errors    # count of errors
    local warnings  # ^^ warnings
    local asserts   # ^^ asserts
    local fails     # ^^ fails
    errors=$(jat__stat perrc)
    warnings=$(jat__stat pwarc)
    asserts=$(jat__stat pasrtc)
    fails=$(jat__stat pfailc)
    case $errors:$warnings:$asserts:$fails in
        0:0:0:*) echo none    ;;
        0:0:*:0) echo pass    ;;
        0:0:*:*) echo fail    ;;
        0:*:*:*) echo warning ;;
        *:*:*:*) echo error   ;;
    esac
}

_jat_core__show_pstart() {
    #
    # Show phase start
    #
    local phint     # phase hint
    phint=$(_jat_core__decorate_hint)
    _jat_core__show \
        "$_JAT_CORE__SWORD_PHASE.$_JAT_CORE__SWORD_START $phint"
}

_jat_core__show_rdata() {
    #
    # Show relic
    #
    local sp=""     # indentation space, if needed
    test "$(_jat_core__rd_keyr phase)" == dummy || sp="  "
    shift   # throw away event type
    _jat_core__show \
        "$sp$_JAT_CORE__SWORD_TEST.$_JAT_CORE__SWORD_RELIC: $*"
}

_jat_core__show_rfile() {
    #
    # Show relic attachment
    #
    local sp=""     # indentation space, if needed
    local path=$2   # file path (rel. to storage)
    local hint=$3   # relic description
    local bracket   # bracket at the end
    test -n "$hint" && bracket=" ($hint)"
    test "$(_jat_core__rd_keyr phase)" == dummy || sp="  "
    _jat_core__show \
        "$sp$_JAT_CORE__SWORD_TEST.$_JAT_CORE__SWORD_RELIC: $path$bracket"
}

_jat_core__show_rfinish() {
    #
    # Show message about run $1 finalization to stderr
    #
    local rverd     # run verdict text
    local errors    # count of errors
    local warnings  # ^^ warnings
    local asserts   # ^^ asserts
    local fails     # ^^ fails
    errors=$(jat__stat rerrc)
    warnings=$(jat__stat rwarc)
    asserts=$(jat__stat rasrtc)
    fails=$(jat__stat rfailc)
    case $errors:$warnings:$asserts:$fails in
        0:0:0:*) rverd=""                            ;;
        0:0:*:0) rverd=".$_JAT_CORE__SWORD_PASS"     ;;
        0:0:*:*) rverd=".$_JAT_CORE__SWORD_FAIL"     ;;
        0:*:*:*) rverd=".$_JAT_CORE__SWORD_TWARNING" ;;
        *:*:*:*) rverd=".$_JAT_CORE__SWORD_TERROR"   ;;
    esac
    _jat_core__show \
        "$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_FINALIZE$rverd"
}

_jat_core__show_rinitn() {
    #
    # Show message about run initialization to stderr
    #
    local items=()
    items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.id: $JAT__TEST_ID")
    items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.version: $JAT__TEST_VERSION")
    test -n "$JAT__TEST_PARAMS" \
     && items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.params: $JAT__TEST_PARAMS")
    test -n "$JAT__TEST_CODE_BASEURL" \
     && items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.code: $(_jat_core__tc_url)")
    test -n "$JAT__TEST_DOC_BASEURL" \
     && items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.doc: $(_jat_core__td_url)")
    items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_START")
    items+=("")
    _jat_core__show "${items[@]}"
}

_jat_core__show_rinitr() {
    #
    # Show message about run reload to stderr
    #
    local items=()
    items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.id: $JAT__TEST_ID")
    items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.version: $JAT__TEST_VERSION")
    test -n "$JAT__TEST_PARAMS" \
     && items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_TINFO: test.params: $JAT__TEST_PARAMS")
    items+=("$_JAT_CORE__SWORD_RUN.$_JAT_CORE__SWORD_RELOAD")
    items+=("")
    _jat_core__show "${items[@]}"
}

_jat_core__show_tinfo() {
    #
    # Show test info
    #
    local sp=""     # indentation space, if needed
    test "$(_jat_core__rd_keyr phase)" == dummy || sp="  "
    _jat_core__show \
        "$sp$_JAT_CORE__SWORD_TEST.$_JAT_CORE__SWORD_TINFO: $1"
}

_jat_core__show_terror() {
    #
    # Show test error
    #
    local sp=""     # indentation space, if needed
    test "$(_jat_core__rd_keyr phase)" == dummy || sp="  "
    _jat_core__show \
        "$sp$_JAT_CORE__SWORD_TEST.$_JAT_CORE__SWORD_TERROR: $1"
}

_jat_core__show_twarning() {
    #
    # Show test warning
    #
    local sp=""     # indentation space, if needed
    test "$(_jat_core__rd_keyr phase)" == dummy || sp="  "
    _jat_core__show \
        "$sp$_JAT_CORE__SWORD_TEST.$_JAT_CORE__SWORD_TWARNING: $1"
}

_jat_core__show_error() {
    #
    # Emit general jat warning (not related to SUT)
    #
    warn "jat.ERROR" "$1"
}

_jat_core__tc_url() {
    #
    # Print test code URL (if $JAT__TEST_CODE_BASEURL is set)
    #
    test -n "$JAT__TEST_CODE_BASEURL" || return 1
    local baseurl=$JAT__TEST_CODE_BASEURL
    local sut_pname
    local version=$JAT__TEST_VERSION
    local relpath
    sut_pname=$(jats__tidparse sut_pname "$JAT__TEST_ID")
    relpath=$(jats__tidparse relpath "$JAT__TEST_ID")
    echo "$baseurl/$sut_pname/$version/$relpath/"
}

_jat_core__td_url() {
    #
    # Print test doc URL (if $JAT__TEST_DOC_BASEURL is set)
    #
    test -n "$JAT__TEST_DOC_BASEURL" || return 1
    local baseurl=$JAT__TEST_DOC_BASEURL
    local sut_pname
    local version=$JAT__TEST_VERSION
    local relpath
    sut_pname=$(jats__tidparse sut_pname "$JAT__TEST_ID")
    relpath=$(jats__tidparse relpath "$JAT__TEST_ID")
    echo "$baseurl/$sut_pname/$version/suite.html#${relpath////-}"
}

__shellfu__jat_core__init() {
    #
    # Initialize '_jat_core' module
    #
    local c_base    # base color (left "border")
    local c_terr    # test error color
    local c_pass    # test pass color
    local c_fail    # test fail color
    local c_info    # info color
    local c_run     # run-related color
    local c_none    # color reset code
    local d_stamps=true     # default stamps presence
    test -t 2 && {
        c_base=${TERMCOLORS_YELLOW}
        c_terr=${TERMCOLORS_LRED}
        c_pass=${TERMCOLORS_GREEN}
        c_fail=${TERMCOLORS_RED}
        c_info=${TERMCOLORS_LBLACK}
        c_run=${TERMCOLORS_LBLUE}
        c_none=${TERMCOLORS_NONE}
        d_stamps=false
    }
    _JAT_CORE__SWORD_ASSERT=${c_base}assert${c_none}
    _JAT_CORE__SWORD_END=${c_info}END${c_none}
    _JAT_CORE__SWORD_FAIL=${c_fail}FAIL${c_none}
    _JAT_CORE__SWORD_FINALIZE=${c_info}FINALIZE${c_none}
    _JAT_CORE__SWORD_PASS=${c_pass}PASS${c_none}
    _JAT_CORE__SWORD_PHASE=${c_base}phase${c_none}
    _JAT_CORE__SWORD_RELOAD=${c_run}RELOAD${c_none}
    _JAT_CORE__SWORD_RELIC=${c_info}RELIC${c_none}
    _JAT_CORE__SWORD_RUN=${c_run}run${c_none}
    _JAT_CORE__SWORD_START=${c_info}START${c_none}
    _JAT_CORE__SWORD_TERROR=${c_terr}ERROR${c_none}
    _JAT_CORE__SWORD_TEST=${c_base}test${c_none}
    _JAT_CORE__SWORD_TINFO=${c_info}INFO${c_none}
    _JAT_CORE__SWORD_TWARNING=${c_terr}WARNING${c_none}
    _JAT_CORE__SCOLOR_INFO=${c_info}
    _JAT_CORE__SCOLOR_NONE=${c_none}
    _JAT_CORE__STAMPS=${_JAT_CORE__STAMPS:-$d_stamps}
}

#shellfu module-version=0.2.12
