#!/bin/bash

shellfu import pretty
shellfu import isa

shellfu import _jat_core
shellfu import _jath

#
# Just A[nother] Test assert and logging library
#
#
# COMMON ARGUMENTS
# ----------------
#
# Assert functions share following parameters:
#
#  *  `-c CASEID`   Case ID; together with $JAT__TEST_ID can be used to
#     uniquely identify a particular test case, with granularity up to
#     single assert.  CASEID set here is joined with any CASEIDs set
#     by active phase.
#
#  *  `-h HINT`   Human-readable description used to describe meaning of
#     the assertion to reader.  Note that HINT should be worded in such
#     way that it makes claim about the reality that matches expectation.
#     That is, to reader of the log, a passing assert tells true, but
#     a failing assert "lies".
#
# Phase start functions share following parameters:
#
#  *  `-C DIR`  Change to directory DIR for duration of the phase.
#
#  *  `-c CASEID`  Inherit CASEID as Case ID to all asserts in this
#     phase.  Individual asserts can extend the final case id for
#     extra granularity and profit.
#
#
# STRINGS
# -------
#
# Many strings, when passed to jat functions, are internally used
# as paths, prefixes or identifiers.  For this reason, they may be
# restricted to contain only "safe" chars.  Following classes are
# defined:
#
#  *  *simple word* - must begin with letter or underscore; second
#     and further characters may be letter, digit or underscore.
#
#
# PARAMETRIZATION
# ---------------
#
# JATS parameters are designed to enable iteration outside control
# of the test script and should be set only by test schedulers.
#
# For example, if a particular test case can only be realized on
# a specific hardware combination, test script can be parametrized
# using this mechanism: scheduler then can provision all necessary
# hardware combinations and schedule the test multiple times, setting
# parameters so that the test code can "know" which mode it's in.
#
# Note that test parameters cannot be used to pass any data; both
# key and value have to be simple words (see STRINGS section of
# this manual.)
#
# When test is started, any JATS parameters must be stored in
# $JAT__TEST_PARAMS.  Test code can access them using jat__param()
# function.
#
#
# EXIT STATUS
# -----------
#
# Most of JAT functions will adhere to following general exit status
# semantics:
#
#  * 0 - everything went OK, or boolean True where applicable.
#  * 1 - boolean False (where applicable, eg. `jat__seen`),
#  * 2 - usage error -- missing arguments, etc.,
#  * 3 - generic error,
#  * 4 - panic.
#
# There are three exceptions:
#
#  *  external code wrappers such as `jat__cmd()` or `jat__eval()`,
#
#  *  `jat__rfinish()`, which normally decides whole test script
#     exit status.
#
#  *  `jat__pend()`, which ca sometimes decide exit status of
#     a wrapper function.
#
#
# ### Code wrappers ###
#
# Call wrapper asserts such as `jat__cmd()` or `jat__eval()` will
# TRY to pass the exit status of the code that was called, but
# MAY return one of the above statuses in case it was not possible
# to call the command in question.
#
# For example, in following code:
#
#     i_will_return_17() { return 17; }
#
#     jat__cmd i_will_return_17             # status will be 17
#     jat__cmd --unkn i_will_return_17      # status will be 2
#
# the first status will be `17` but the second one will be `2` because
# `jat__cmd()` will have failed to understand the `--unkn` argument.
#
# Note that when argument parsing succeeds, jat__cmd() and
# jat__eval() also perform validity check on the status of the
# wrapped code.  This check (see JAT__CMD_VALID_ES for details)
# does not affect exit status of the assert function.
#
#
# ### jat__rfinish() and script exit status ###
#
# Second exception to the generic status semantics is `jat__rfinish()`,
# which is designed to be at the end of the script and represent status
# of the whole test run.
#
# This overall status is computed based on how many FAIL, WARNING or
# ERROR events were collected in this run.  The logic goes this way:
#
#  *  0 - No FAIL's, WARNING's or ERROR's.
#
#  *  1 - At least one FAIL, but no WARNING's or ERROR's.
#
#  *  2 - At least one WARNING, but no ERROR's.
#
#  *  3 - This status is reserved.
#
#  *  4 - At least one ERROR.
#
# Note that the semantic is similar to the TFKit semantic:
#
#       .---------------------------------------------------------------.
#       | e |    state of         |                                     |
#       | s |---------------------| script says                         |
#       |   | SUT   | environment |                                     |
#       |---|-------|-------------|-------------------------------------|
#       | 0 | OK    | safe        | test passed, everything worked fine |
#       | 1 | buggy | safe        | test failed, everything worked fine |
#       | 2 | ???   | safe        | Something made me skip a test       |
#       | 3 | ???   | safe        | Something blew up but I managed to  |
#       |   |       |             | clean up (I promise!)               |
#       | 4 | ???   | broken      | Something blew up and I rushed out  |
#       |   |       |             | in panic                            |
#       | * | ???   | broken      | ...nothing (is dead)                |
#       '---------------------------------------------------------------'
#
# except that JAT has no standard way of distinguishing between
# `2` and `3`, therefore `3` is not used by JAT.
#
# See also TFKit documentation:
#
#     https://gitlab.com/vornet/infra/tfkit/-/tree/master/src/doc#on-exit-statuses-three-and-above
#
#
# ### jat__pend() and function exit status ###
#
# It's common in JATS tests to wrap phases in functions like this:
#
#     assert_something() {
#         #
#         # Assert that something exists
#         #
#         jat__pstartt -h "something exists"
#             jat__cmd test -f /etc/something
#         jat__pend
#     }
#
#     assert_something || jat__abort "no something"
#
# In those situations, the exit status of jat__pend() will affect
# the function exit status, so the above section about
# jat__rfinish() mostly applies to jat__pend() as well.
#

#
# Valid exit status expression
#
# Comma-separated list of exit statuses or exit status ranges,
# defining valid exit statuses for jat__cmd() and jat__eval().
#
# Exit statuses that don't match neither expected exit status
# of given assert nor this expression will raise warning.
#
# For example, following set of asserts will produce three
# PASSes, two FAILs and one WARNING:
#
#     jat__cmd      test 6 -eq 6    # PASS - `test` gave 0
#     jat__cmd      test 6 -eq 7    # FAIL - `test` gave 1
#     jat__cmd -S 1 test 6 -eq 7    # PASS - `test` gave 1
#     jat__cmd      test 6 -eq six  # FAIL + WARNING - `test` gave 2
#     jat__cmd -S 2 test 6 -eq six  # PASS - `test` gave 2
#
# Notice that the last assert did not produce WARNING as the
# `2` was explicitly expected.
#
# The use case for JAT__CMD_VALID_ES is when we don't expect
# the status but (perhaps if the command is likely to produce
# various statuses on failure) still don't want to treat
# "weird" status as suspicious:
#
#     JAT__CMD_VALID_ES=0-255 \
#       jat__cmd      test 6 -eq six  # FAIL - `test` gave 2
#
# Note that this validation does not affect exit status of the
# assert function itself; for details on that, see EXIT STATUS
# section of this manual.
#
JAT__CMD_VALID_ES=${JAT__CMD_VALID_ES:-0,1}

#
# Path to result directory
#
JAT__DIR=${JAT__DIR:-/var/tmp/jat}

#
# Path to result log
#
# If unset, another identical log is *still* created under a unique
# subdirectory of $JAT__DIR.  Set this if you want to have log at
# a predictable path.  This can be useful  eg. if you want to send
# log into a named pipe.
#
JAT__YLOG=${JAT__YLOG:-}

#
# Test name
#
JAT__TEST_ID=${JAT__TEST_ID:-_jat_anon_test_}

#
# Test parameters
#
# Comma-separated list of key/value pairs.  These are supposed
# to be passed by scheduler and can be accessed from within the test
# using jat__param() and jat__has_param().
#
# Note that *both* key and value have to be simple words.  See also
# "PARAMETRIZATION" section of this manual.
#
JAT__TEST_PARAMS=${JAT__TEST_PARAMS:-}

#
# Test version
#
JAT__TEST_VERSION=${JAT__TEST_VERSION:-_jat_no_test_version_}

#
# Ranges of YAML-invalid characters
#
# List of Perl regex ranges; must be usable in both s/[]// and tr///.
# Note that by default this list is set during shellfu init unless
# it has been set to non-empty string.
#
# Overriding might result in invalid YAML log!
#
__JAT__YAML_BAD=${__JAT__YAML_BAD:-}

#
# Replacement character for YAML-invalid strings
#
# Normal string data should be already quoted based on $__JAT__YAML_BAD
# ranges.  However, if something slips through (either by bug in this
# library or as part of eg. field name), there's one extra pass that
# replaces all prohibited characters by this one.
#
__JAT__YAML_REPL=${__JAT__YAML_REPL:-�}

#
# Make result log deterministic?
#
# Determines whether result should be made deterministic.  Currently
# this means that instead of wall clock time, timestamps will be simply
# ordinal stamps like `time-1`, `time-2`...
#
# This makes test log deterministic, enabling comparison of multiple
# runs against each other or against fixed oracle.  The disadvantage
# is that you can't assess performance data from the log.
#
__JAT__DETERMINISTIC=${__JAT__DETERMINISTIC:-false}

#
# Log format version
#
__JAT__LOG_FMT='jat/0.0'

#
# Self version (in variable to enable overriding in unit tests)
#
__JAT__SELF_VERSION=0.2.12

jat__abort() {
    #
    # Exit this script abruptly for reason $1 with details $2..
    #
    # Usage:
    #     jat__abort REASON [--] [LINE]..
    #     jat__abort REASON -f [FILE]..
    #     jat__abort REASON -c CMD [ARG]..
    #     jat__abort
    #
    # REASON it must be a short one-line string summarizing the reason
    # why the test script is aborting.  REASON may be supplemented with
    # further detail report, which can be added by providing individual
    # LINEs, path to FILE that could be concatenated or CMD that would
    # produce the report upon calling.  If REASON is omoitted, string
    # '_jat_no_reason_' is provided by default.
    #
    # Before reporting error and exiting, this function will create and
    # submit file 'jat__abort.reason', which has following structure:
    #
    #     REASON
    #
    #     any further
    #     lines of
    #     detailed report
    #
    # That is, REASON, followed by an empty line, followed by any further
    # lines generated by LINE, CMD or FILE arguments.
    #
    # Note that REASON is intended to not just describe the reason for
    # human reader, but also categorize the failure mode for purposes
    # of further processing such as metrics.  That is, REASON should
    # contain enough information to discern between different failure
    # modes but not too much to make it non-deterministic.
    #
    # For example, if reason of aborting the test is a setup failure
    # caused by dependency errors, good REASON could be `dependency error`
    # or `dependency error: perl`.  On the other hand, `dependency error:
    # perl-5.27-34` might be too granular for useful metrics; details such
    # as package versions should go to the detailed report instead.
    #
    local reason=${1:-_jat_no_reason_}; shift
    local tmp
    local src=none
    local args=()
    while true; do case $1 in
        -c) src="cmd";  shift; break ;;
        -f) src="file"; shift; break ;;
        --) src="line"; shift; break ;;
        "") break ;;
        *)  src="line"; break ;;
    esac done
    args=("$@")
    tmp=$(mktemp -t jat__abort.XXXXXXXX)
    (
        echo "$reason"
        test "$src" == "none" && return
        echo
        test ${#args[@]} -gt 0 || __jat__usage "no ${src^^}?"
        case $src in
            cmd)  "${args[@]}"               ;;
            file) cat "${args[@]}"           ;;
            line) printf '%s\n' "${args[@]}" ;;
        esac
    )>"$tmp"
    jat__submit "$tmp" "jat__abort.reason"
    rm "$tmp"
    jat__die "aborting: $reason"
}

jat__cmd() {
    #
    # Assert that command $@ exits with zero (or other)
    #
    # Usage:
    #
    #     jat__cmd [assert-options] [-o R_OUT] [-e R_ERR] [-s R_ESF] [-S ES_EXPR] [-V ES_EXPR] [--] CMD [ARG]..
    #
    # Run CMD with all ARGs and if exit status is zero, announce assertion
    # success (ASSERT.PASS), otherwise announce assertion failure (ASSERT.FAIL).
    #
    # Exit status expectation can be changed from zero to ES_EXPR, which
    # must be in form of comma-separated list of integer values or ranges
    # thereof.  E, g. expression '0,10-20' would be a valid expression,
    # matching 0, 11, 20, but not 2 or 21.
    #
    # Exit status is also matched for validity, which means that if exit
    # status matches neither value of -S nor value of $JAT__CMD_VALID_ES
    # environment variable, a warning is raised.  This can be overriden
    # using -V parameter.
    #
    # If R_ESF is passed, it will be treated as path to file where exit status
    # of CMD is written.  This is recommended over consulting $?, since the
    # latter is not reliable; see below.
    #
    # If path R_OUT or R_ERR are provided, standard output and standard
    # error are saved to respective files.
    #
    # See COMMON ARGUMENTS section for common assert options.
    #
    # Exit status of the this function normally will, but IS NOT GUARRANTEED TO
    # correspond to the exit status of the CMD. For example, if the function
    # call is incomplete, the exit status will be 2.
    #
    local __jat__cmd=()     # CMD with all ARGs
    local __jat__r_es       # result: exit status
    local __jat__o_es=0     # oracle: exit status
    local __jat__t_ves="$JAT__CMD_VALID_ES"  # test: valid exit statuses
    local __jat__hint       # assert HINT
    local __jat__caseid     # assert CASEID
    local __jat__etype=MESSAGE.ERROR    # event type
    local __jat__r_out      # result: stdout
    local __jat__r_err      # result: stderr
    local __jat__r_esf      # result: exit status file
    local __jat__tdur       # duration of running command
    #
    # NOTE: names need to be qualified because they might interfere
    #       with the actual execution of the assert command.
    #
    while true; do case $1 in
        -c) __jat__caseid="$2"; shift 2 || { __jat__usage "missing CASEID"; return 2; } ;;
        -h) __jat__hint=$2; shift 2 || { __jat__usage "missing HINT"; return 2; } ;;
        -e) __jat__r_err=$2; shift 2 || { __jat__usage "missing R_ERR"; return 2; } ;;
        -o) __jat__r_out=$2; shift 2 || { __jat__usage "missing R_OUT"; return 2; } ;;
        -s) __jat__r_esf=$2; shift 2 || { __jat__usage "missing R_ESF"; return 2; } ;;
        -S) __jat__o_es=$2; shift 2 || { __jat__usage "missing ES_EXPR"; return 2; } ;;
        -V) __jat__t_ves=$2; shift 2 || { __jat__usage "missing ES_EXPR"; return 2; } ;;
        --) shift; break ;;
        -*) __jat__usage "unknown argument: $1"; return 2 ;;
        *) break ;;
    esac done
    __jat__cmd=("$@")
    test -n "${__jat__cmd[*]}" || { __jat__usage "no CMD?"; return 2; }
    test -n "$__jat__hint" || __jat__hint="${__jat__cmd[*]}"
    debug -v __jat__cmd __jat__o_es __jat__hint
    __jat__tic jat__cmd
    case ${#__jat__r_out}:${#__jat__r_err} in
        0:0) "${__jat__cmd[@]}" ;;
        0:*) "${__jat__cmd[@]}" 2>"$__jat__r_err" ;;
        *:0) "${__jat__cmd[@]}" >"$__jat__r_out" ;;
        *:*) "${__jat__cmd[@]}" >"$__jat__r_out" 2>"$__jat__r_err" ;;
    esac; __jat__r_es=$?
    __jat__tdur=$(__jat__toc jat__cmd)
    debug -v __jat__r_es __jat__r_esf
    if test -n "$__jat__r_esf"; then
        echo $__jat__r_es > "$__jat__r_esf" \
         || jat__log_error "error writing R_ESF file: $__jat__r_esf"
    fi
    if __jat__es_match "$__jat__o_es" "$__jat__r_es"; then
        __jat__etype=ASSERT.PASS
    else
        __jat__etype=ASSERT.FAIL
        _jat_core__show_adetail "exit status did not match: $__jat__r_es !~ $__jat__o_es" \
            "exit status oracle: $__jat__o_es" \
            "exit status: $__jat__r_es" \
            "duration: $__jat__tdur" \
            "command (unquoted here): ${__jat__cmd[*]}"
        __jat__es_match "$__jat__t_ves" "$__jat__r_es" \
         || jat__log_warning \
                "invalid exit status: $__jat__r_es (not $__jat__t_ves)" \
                "(hint: use -V or set JAT__CMD_VALID_ES)"
    fi
    __jat__assert $__jat__etype "$__jat__hint" "$__jat__caseid" \
        "t.cmd=${__jat__cmd[*]}" "o.es_expr=$__jat__o_es" "r.es=$__jat__r_es" \
        "r.duration=$__jat__tdur"
    return "$__jat__r_es"
}

jat__cmp() {
    #
    # Assert comparison of $1 and $3 using operator $2
    #
    # Usage:
    #
    #     jat__cmp [assert-options] RVAL OP OVAL
    #
    # Compare RVAL (result value) to OVAL (oracle value) and announce
    # assertion success (ASSERT.PASS) or assertion failure (ASSERT.FAIL).
    #
    # OP can be any of `eq`, `ne`, `lt`, `gt`, `le`, `ge` for numeric values,
    # `==` and `re` for string values.  In case of `re`, OVAL is expected to
    # be basic regular expression and is matched against whole string (ie.
    # anchors `^` and `$` are included automatically.
    #
    # Note that in most cases using jat__cmd() with `test` or `grep` commands
    # is all you need.
    #
    # See COMMON ARGUMENTS section for common assert options.
    #
    local RVal      # "result" value
    local Op        # operator
    local OVal      # "oracle" value
    local hint      # log hint
    local caseid    # case id
    local cmpes     # comparison exit status
    while true; do case $1 in
        -c) caseid="$2"; shift 2 || { __jat__usage "missing CASEID"; return 2; } ;;
        -h) hint=$2; shift 2 || { __jat__usage "missing HINT"; return 2; } ;;
        *) break ;;
    esac done
    RVal=$1; Op=$2; OVal=$3
    test $# -eq 0 && { __jat__usage "no RVAL?"; return 2; }
    test $# -eq 1 && { __jat__usage "no OP?"; return 2; }
    test $# -eq 2 && { __jat__usage "no OVAL?"; return 2; }
    test $# -gt 3 && { __jat__usage "too many arguments!"; return 2; }
    test -n "$hint" || hint="$RVal $Op $OVal"
    debug -v RVal Op OVal hint
    __jat__cmp_match; cmpes=$?
    case $cmpes in
        0)
            __jat__assert ASSERT.PASS "$hint" "$caseid" \
                "r.val=$RVal" "t.op=$Op" "o.val=$OVal"
            ;;
        1)
            _jat_core__show_adetail "no match: $RVal $Op $OVal" \
                "result value: $RVal" \
                "test operator: $Op" \
                "oracle value: $OVal"
            __jat__assert ASSERT.FAIL "$hint" "$caseid" \
                "r.val=$RVal" "t.op=$Op" "o.val=$OVal"
            ;;
        *)
            __jat__usage "bad syntax: $RVal $Op $OVal"
            return 2
            ;;
    esac
}

jat__die() {
    #
    # Log error $1 and exit this script abruptly
    #
    local msg
    local msgs=()
    for msg in "$@"; do
        msgs+=("fatal: $msg")
    done
    jat__log_error "${msgs[@]}"
    _jat_core__rd_keyw runstatus aborted
    __jat__rfinish; es=$?
    warn "${msgs[@]}"
    exit $es
}

jat__eval() {
    #
    # Assert that code $1 exits with zero (or other)
    #
    # Usage:
    #
    #     jat__eval [assert-options] [-o R_OUT] [-e R_ERR] [-s R_ESF] [-S ES_EXPR] [-V ES_EXPR] [--] CODE
    #
    # Run CODE using eval builtin and if exit status is zero, announce assertion
    # success (ASSERT.PASS), otherwise announce assertion failure (ASSERT.FAIL).
    #
    # Exit status expectation can be changed from zero to ES_EXPR, which
    # must be in form of comma-separated list of integer values or ranges
    # thereof.  E, g. expression '0,10-20' would be a valid expression,
    # matching 0, 11, 20, but not 2 or 21.
    #
    # Exit status is also matched for validity, which means that if exit
    # status matches neither value of -S nor value of $JAT__CMD_VALID_ES
    # environment variable, a warning is raised.  This can be overriden
    # using -V parameter.
    #
    # If R_ESF is passed, it will be treated as path to file where exit status
    # of CODE is written.  This is recommended over consulting $?, since the
    # latter is not reliable; see below.
    #
    # If path R_OUT or R_ERR are provided, standard output and standard
    # error are saved to respective files.
    #
    # This is similar to jat__cmd(), except that code is checked against syntax
    # errors and user is given full control over redirections.
    #
    # See COMMON ARGUMENTS section for common assert options.
    #
    # Exit status of the this function normally will, but IS NOT GUARRANTEED TO
    # correspond to the exit status of the CODE. For example, if the function
    # call is incomplete, the exit status will be 2.  If there is syntax error
    # in CODE, the exit status will be 3.
    #
    local __jat__caseid     # assert CASEID
    local __jat__code=""    # CODE to evaluate
    local __jat__r_es       # result: exit status
    local __jat__t_ves="$JAT__CMD_VALID_ES"  # test: valid exit statuses
    local __jat__r_out      # result: stdout
    local __jat__r_err      # result: stderr
    local __jat__r_esf      # result: exit status file
    local __jat__o_es=0     # oracle: exit status
    local __jat__hint       # assert HINT
    local __jat__etype=MESSAGE.ERROR    # event type
    local __jat__tdur       # duration of running code
    #
    # NOTE: names need to be qualified because they might interfere
    #       with the actual execution of the assert command.
    #
    while true; do case $1 in
        -c) __jat__caseid="$2"; shift 2 || { __jat__usage "missing CASEID"; return 2; } ;;
        -e) __jat__r_err=$2; shift 2 || { __jat__usage "missing R_ERR"; return 2; } ;;
        -h) __jat__hint=$2; shift 2 || { __jat__usage "missing HINT"; return 2; } ;;
        -o) __jat__r_out=$2; shift 2 || { __jat__usage "missing R_OUT"; return 2; } ;;
        -s) __jat__r_esf=$2; shift 2 || { __jat__usage "missing R_ESF"; return 2; } ;;
        -S) __jat__o_es=$2; shift 2 || { __jat__usage "missing ES_EXPR"; return 2; } ;;
        -V) __jat__t_ves=$2; shift 2 || { __jat__usage "missing ES_EXPR"; return 2; } ;;
        --) shift; break ;;
        -*) __jat__usage "unknown argument: $1"; return 2 ;;
        *) break ;;
    esac done
    __jat__code=$1
    test -n "$__jat__code" || { __jat__usage "no CODE?"; return 2; }
    test -n "$2" &&  { __jat__usage "extra args!: $*"; return 2; }
    test -n "$__jat__hint" || __jat__hint="$__jat__code"
    debug -v __jat__code __jat__o_es __jat__hint
    bash -n <<<"$__jat__code" || {
        jat__log_error "syntax error in test code: $__jat__code"
        return 3
    }
    __jat__tic jat__eval
    case ${#__jat__r_out}:${#__jat__r_err} in
        0:0) ( eval "$__jat__code" ) ;;
        0:*) ( eval "$__jat__code" ) 2>"$__jat__r_err" ;;
        *:0) ( eval "$__jat__code" ) >"$__jat__r_out" ;;
        *:*) ( eval "$__jat__code" ) >"$__jat__r_out" 2>"$__jat__r_err" ;;
    esac; __jat__r_es=$?
    __jat__tdur=$(__jat__toc jat__eval)
    if test -n "$__jat__r_esf"; then
        echo $__jat__r_es > "$__jat__r_esf" \
         || jat__log_error "error writing R_ESF file: $__jat__r_esf"
    fi
    debug -v __jat__r_es
    if __jat__es_match "$__jat__o_es" "$__jat__r_es"; then
        __jat__etype=ASSERT.PASS
    else
        _jat_core__show_adetail "exit status did not match: $__jat__r_es !~ $__jat__o_es" \
            "exit status oracle: $__jat__o_es" \
            "exit status: $__jat__r_es" \
            "duration: $__jat__tdur" \
            "code (unquoted here): $__jat__code"
        __jat__es_match "$__jat__t_ves" "$__jat__r_es" \
         || jat__log_warning \
                "invalid exit status: $__jat__r_es (not $__jat__t_ves)" \
                "(hint: use -V or set JAT__CMD_VALID_ES)"
        __jat__etype=ASSERT.FAIL
    fi
    __jat__assert $__jat__etype "$__jat__hint" "$__jat__caseid" \
        "t.code=$__jat__code" "o.es_expr=$__jat__o_es" "r.es=$__jat__r_es" \
        "r.duration=$__jat__tdur"
    return "$__jat__r_es"
}

jat__fail() {
    #
    # Assert test failure
    #
    # Usage:
    #
    #     jat__fail [-d DKEY=DVALUE] [assert-options]
    #
    # Simply log event that an assert has failed.  Use other assert functions
    # if you can, since they probably provide more useful information to log
    # reader.
    #
    # See COMMON ARGUMENTS section for common assert options.
    #
    # Additionally, you can pass extra information (typically the particular
    # piece that decided the assert) by adding `-d` argument followed by
    # a key/value pair.  You cam add `-d DKEY=DVALUE` multiple times to
    # create a "dictionary" of values.
    #
    local hint      # log hint
    local caseid    # case ID
    local data=()   # assert data
    while true; do case $1 in
        -c) caseid="$2"; shift 2 || { __jat__usage "missing CASEID"; return 2; } ;;
        -d) data+=("$2"); shift 2 || { __jat__usage "missing K=V"; return 2; } ;;
        -h) hint=$2; shift 2 || { __jat__usage "missing HINT"; return 2; } ;;
        *) break ;;
    esac done
    test -n "$1" && { __jat__usage "extra arguments: $*"; return 2; }
    debug -v hint
    _jat_core__show_adetail "assert fail"
    __jat__assert ASSERT.FAIL "$hint" "$caseid" "${data[@]}"
}

jat__filebackup() {
    #
    # Back up file $1
    #
    # Usage:
    #
    #     jat__filebackup [-n NS] [-c] [--] PATH
    #
    # Back up PATH aside to be able to restore it later using
    # jat__filerestore().
    #
    # Provide `-c` switch to ensure that path is removed before restoring.
    # If NS is given, it must be "simple word" (see STRINGS section) and
    # will cause jat__filebackup() to use separate srorage for backups.
    #
    local path      # path to back up
    local abspath   # absolute path ^^
    local ns=_jat_anon_backupns_    # this backup's namespace
    local store     # this backup's store dir
    local dest      # final path to backed up data within store
    local clean=false
    local pdig      # path DIGEST from rainbow file
    local nshint    # namespace hint in messages
    local uselog    # use log (list backup DIGESTs ever restored)
    local ovhint    # overwriting hint in backup message
    while true; do case $1 in
        --) shift; break ;;
        -c|--clean)
            clean=true
            shift ;;
        -n|--namespace)
            ns=$2; nshint="(NS=$ns) "
            shift 2 || { __jat__usage "no NS?"; return 2; }
            ;;
        -*) __jat__usage "unknown argument: $1"; return 2 ;;
        *) break ;;
    esac done
    path=$1
    test -n "$path" || { __jat__usage "no PATH?"; return 2; }
    isa__name "$ns" || { __jat__usage "NS must be simple word"; return 2; }
    abspath=$(readlink -e "$path") || {
        jat__log_error "no such file: $1"
        return 3
    }
    pdig=$(md5sum <<<"$abspath" | cut -d' ' -f1)
    _jat_core__rd_keyA "backup/$ns/rainbow" "$pdig $abspath"
    $clean && _jat_core__rd_keya "backup/$ns/cleanlog" "$abspath"
    uselog=$(_jat_core__rd_path "backup/$ns/uselog")
    store=$(_jat_core__rd_path "backup/$ns/store")
    test -n "$store" || {
        jat__log_error "bad storage path: store='$store'"
        return 4
    }
    mkdir -p "$store"
    dest="$store/$pdig"
    test -e "$store/$pdig" && {
        grep -sqxFe "$pdig" "$uselog" \
         || jat__log_warning "overwriting unused backup: $nshint$path"
        #shellcheck disable=SC2115
        rm -r "$store/$pdig"
        ovhint=" (overwrote)"
    }
    cp -ar -- "$abspath" "$store/$pdig" || {
        jat__log_error "failed to create backup: $abspath"
        return 3
    }
    jat__log_info "backed up$ovhint: $nshint$path"
}

jat__filerestore() {
    #
    # Restore paths stored by jat__filebackup()
    #
    # Usage:
    #
    #     jat__filerestore [-s] [-n NS]
    #
    # Restore all paths backed up using jat__filebackup().  Paths
    # are restored in opposite order than backed up.
    #
    # If you used NS with jat__filebackup(), provide the same string
    # with this function as well.
    #
    # By default, attempting to restore files from non-existent NS
    # will log ERROR and return with exit status 3.  Use -s switch to
    # silence the error, ie. jat__filerestore() will log INFO and
    # return with zero.
    #
    local abspath   # ORIGPATH from rainbow file
    local ns=_jat_anon_backupns_    # this backup's namespace
    local store     # this backup's store dir
    local pdig      # path DIGEST from rainbow file
    local nshint    # namespace hint in messages
    local rainbow   # path to rainbow file (lines of 'DIGEST ORIGPATH'
    local cleanlog  # clean log (if ORIGPATH is there we shold pre-clean it)
    local uselog    # use log (list backup DIGESTs ever restored)
    local silent=false   # silence NS errors
    while true; do case $1 in
        -n|--namespace)
            ns=$2; nshint="(NS=$ns) "
            shift 2 || { __jat__usage "no NS?"; return 2; }
            ;;
        -s|--silent)
            silent=true; shift
            ;;
        "") break ;;
        *) __jat__usage "unknown argument: $1"; return 2 ;;
    esac done
    isa__name "$ns" || { __jat__usage "NS must be simple word"; return 2; }
    test -d "$_JAT_CORE__RDIR/internal/backup/$ns" || {
        if $silent; then
            jat__log_info "nothing to restore from: $ns"
            return 0
        else
            jat__log_error "unknown backup namespace: $ns"
            return 4
        fi
    }
    rainbow=$(_jat_core__rd_path "backup/$ns/rainbow")
    cleanlog=$(_jat_core__rd_path "backup/$ns/cleanlog")
    store=$(_jat_core__rd_path "backup/$ns/store")
    uselog=$(_jat_core__rd_path "backup/$ns/uselog")
    while read -r pdig abspath; do
        debug -v pdig abspath
        grep -qxFe "$abspath" "$cleanlog" 2>/dev/null && {
            rm -rf "$abspath" || {
                jat__log_error "failed to pre-clean original path: $abspath"
                return 4
            }
        }
        cp -Tar -- "$store/$pdig" "$abspath" || {
            jat__log_error "failed to restore backup: $nshint$abspath"
            return 4
        }
        echo "$pdig" >> "$uselog"
        jat__log_info "restored: $nshint$abspath"
        :
    done <"$rainbow"
}

jat__rfinish() {
    #
    # Finalize run
    #
    # Usage;
    #     jat__rfinish
    #
    # Log event to announce that log is finalized and remove test
    # run.
    #
    # This is necessary because in order to support persistent multi
    # PID tests (ie, re-booting within test), jat__rinit() will reload
    # leftover run and whole log will be merged.
    #
    test $# -gt 0 && { __jat__usage "extra arguments: $*"; return 5; }
    #                2 has other meaning for this function ^^^
    test -d "$JAT__DIR/run" || {
        _jat_core__show_error "no active run: no $JAT__DIR/run"
        return 5
        # ^^^ 2 has other meaning for this function
    }
    _jat_core__rd_keyw runstatus completed
    __jat__rfinish
}

__jat__rfinish() {
    #
    # Finalize run
    #
    local fileas    # run id for filing in 'finished' folder
    local es        # final exit status
    es=$(__jat__final_es)
    _jat_core__show_rfinish
    __jat__log_event RUN.END
    __jat__writelog <<<"finalized: true"
    __jat__writelog <<<"end: $(__jat__newstamp)"
    _jath__runstatus "$(_jat_core__rd_keyr runstatus)"
    fileas="run-$(_jat_core__rd_keyr runid)"
    mkdir -p "$JAT__DIR/finished"
    mv "$JAT__DIR/run" "$JAT__DIR/finished/$fileas"
    rm -f "$JAT__DIR/last"
    ln -s "finished/$fileas" "$JAT__DIR/last"
    return "$es"
}

jat__log_error() {
    #
    # Log internal error message of lines $@
    #
    # Usage:
    #     jat__log_error [LINE]..
    #
    # Create error message (MESSAGE.ERROR event in YAML log) consisting
    # of provided LINEs.
    #
    __jat__log error "$@"
}

jat__log_errorv() {
    #
    # Log contents of internal variables named $@ as ERROR
    #
    # Usage:
    #     jat__log_infov VARNAME [VARNAME..]
    #
    # Creates log message showing current values of all given variables,
    # or string `#Unset` for undefined variables.
    #
    # Every VARNAME must be a valid variable name according to rules
    # described as "simple words" in STRINGS section of this manual.
    #
    __jat__logv error "$@"
}

jat__log_info() {
    #
    # Log internal info message of lines $@
    #
    # Usage:
    #     jat__log_info [LINE]..
    #
    # Create info message (MESSAGE.INFO event in YAML log) consisting
    # of provided LINEs.
    #
    __jat__log info "$@"
}

jat__log_infov() {
    #
    # Log contents of internal variables named $@
    #
    # Usage:
    #     jat__log_infov VARNAME [VARNAME..]
    #
    # Creates log message showing current values of all given variables,
    # or string `#Unset` for undefined variables.
    #
    # Every VARNAME must be a valid variable name according to rules
    # described as "simple words" in STRINGS section of this manual.
    #
    __jat__logv info "$@"
}

jat__log_warning() {
    #
    # Log internal info message of lines $@
    #
    # Usage:
    #     jat__log_warning [LINE]..
    #
    # Create info message (MESSAGE.WARNING event in YAML log) consisting
    # of provided LINEs.
    #
    __jat__log warning "$@"
}

jat__log_warningv() {
    #
    # Log contents of internal variables named $@ as WARNING
    #
    # Usage:
    #     jat__log_infov VARNAME [VARNAME..]
    #
    # Creates log message showing current values of all given variables,
    # or string `#Unset` for undefined variables.
    #
    # Every VARNAME must be a valid variable name according to rules
    # described as "simple words" in STRINGS section of this manual.
    #
    __jat__logv warning "$@"
}

jat__pass() {
    #
    # Assert test pass
    #
    # Usage:
    #
    #     jat__pass [-d DKEY=DVALUE] [assert-options]
    #
    # Simply log event that an assert has passed.  Use other assert functions
    # if you can, since they probably provide more useful information to log
    # reader.
    #
    # See COMMON ARGUMENTS section for common assert options.
    #
    # Additionally, you can pass extra information (typically the particular
    # piece that decided the assert) by adding `-d` argument followed by
    # a key/value pair.  You cam add `-d DKEY=DVALUE` multiple times to
    # create a "dictionary" of values.
    #
    local hint      # log hint
    local caseid    # case ID
    local data=()   # assert data
    while true; do case $1 in
        -c) caseid="$2"; shift 2 || { __jat__usage "missing CASEID"; return 2; } ;;
        -d) data+=("$2"); shift 2 || { __jat__usage "missing K=V"; return 2; } ;;
        -h) hint=$2; shift 2 || { __jat__usage "missing HINT"; return 2; } ;;
        *) break ;;
    esac done
    test -n "$1" && { __jat__usage "extra arguments: $*"; return 2; }
    debug -v hint
    __jat__assert ASSERT.PASS "$hint" "$caseid" "${data[@]}"
}

jat__promise_asserts() {
    #
    # Promise number of asserts will be $1
    #
    local num=$1    # promised number
    test -n "$num" || {
        __jat__usage "no NUM?"
        return 2
    }
    __jat__log_event TEST.PROMISE "assert number: $num" "" \
        "num=$num"
}

jat__rinit() {
    #
    # Initialize run
    #
    # Usage:
    #     jat__rinit
    #
    # Load active run if found; otherwise initialize new one.
    #
    local reload=false      # 'true' if reloading old run
    local old_test_id       # old test id if ^^
    local old_test_version  # old test version if ^^
    local old_test_params   # old test params if ^^
    local runid            # this run id
    test $# -gt 0 && { __jat__usage "extra arguments: $*"; return 2; }
    mkdir -p "$JAT__DIR" \
     || die "could not initialize JAT__DIR: $JAT__DIR"
    _JAT_CORE__RDIR=$JAT__DIR/run
    test -d "$_JAT_CORE__RDIR" && reload=true
    mkdir -p "$_JAT_CORE__RDIR"
    export _JAT_CORE__RDIR
    export JAT__YLOG
    debug -v reload _JAT_CORE__RDIR JAT__TEST_PARAMS
    if $reload; then
        runid=$(_jat_core__rd_keyr runid)
        old_test_id=$(_jat_core__rd_keyr test_id)
        old_test_version=$(_jat_core__rd_keyr test_version)
        old_test_params=$(_jat_core__rd_keyr test_params)
        test "$old_test_id" == "$JAT__TEST_ID" || {
            warn -v JAT__TEST_ID old_test_id
            die "cannot reload run of different test"
        }
        test "$old_test_version" == "$JAT__TEST_VERSION" || {
            warn -v JAT__TEST_VERSION old_test_version
            die "cannot reload run of different test version"
        }
        if test -n "$JAT__TEST_PARAMS"; then
            test "$old_test_params" == "$JAT__TEST_PARAMS" || {
                warn -v JAT__TEST_PARAMS old_test_params
                die "cannot reload run of different test params"
            }
        fi
        JAT__TEST_PARAMS=$old_test_params
        _jat_core__rd_keyw runstatus running
        __jat__log_event RUN.RELOAD
        _jat_core__show_rinitr
    else
        runid=$(__jat__newsid)
        debug -v _JAT_CORE__RDIR JAT__YLOG
        {
            echo "---"
            echo "format: $__JAT__LOG_FMT"
            echo "jat_version: $__JAT__SELF_VERSION"
            echo "test:"
            echo "    id: $JAT__TEST_ID"
            echo "    version: $JAT__TEST_VERSION"
            test -n "$JAT__TEST_PARAMS" && echo "    params: $JAT__TEST_PARAMS"
            test -n "$JAT__TEST_CODE_BASEURL" \
             && echo "    code: $(_jat_core__tc_url)"
            test -n "$JAT__TEST_DOC_BASEURL" \
             && echo "    doc: $(_jat_core__td_url)"
            echo "start: $(__jat__newstamp)"
            echo "id: $runid"
            echo "events:"
        } | __jat__writelog
        _jat_core__rd_keyw runid "$runid"
        _jat_core__rd_keyw test_id "$JAT__TEST_ID"
        _jat_core__rd_keyw test_version "$JAT__TEST_VERSION"
        _jat_core__rd_keyw test_params "$JAT__TEST_PARAMS"
        __jat__pdummy
        __jat__log_event RUN.START
        _jat_core__rd_keyw runstatus running
        _jat_core__show_rinitn
    fi
    _jath__runstatus "$(_jat_core__rd_keyr runstatus)"
    true
}

jat__seen() {
    #
    # True if any $1 were seen
    #
    # Usage:
    #    jat__seen rasrt       # asserts since run start
    #    jat__seen rpass       # passes since run start
    #    jat__seen rfail       # fails since run start
    #    jat__seen pasrt       # asserts since phase start
    #    jat__seen ppass       # passes since phase start
    #    jat__seen pfail       # fails since phase start
    #    jat__seen rwar        # warnings since run start
    #    jat__seen pwar        # warnings since phase start
    #    jat__seen rerr        # errors since run start
    #    jat__seen perr        # errors since phase start
    #
    # Check important occurences in run state.
    #
    local which=$1
    case $which in
        rwar|pwar|rerr|perr|rfail|pfail|rpass|ppass|rasrt|pasrt)
            : ;;
        *) __jat__usage "invalid statistic field: $which"
           return 2 ;;
    esac
    test "$(jat__stat "${which}c")" -gt 0
}

jat__stat() {
    #
    # Print run statistic $1
    #
    # Usage:
    #    jat__stat rasrtc      # run assert count
    #    jat__stat rpassc      # run pass count
    #    jat__stat rfailc      # run fail count
    #    jat__stat pasrtc      # phase assert count
    #    jat__stat ppassc      # phase pass count
    #    jat__stat pfailc      # phase fail count
    #    jat__stat rwarc       # run warning count
    #    jat__stat pwarc       # phase warning count
    #    jat__stat rerrc       # run error count
    #    jat__stat perrc       # phase error count
    #
    # Print statistics about run state.
    #
    local which=$1  # statistic key
    case $which in
        rwarc)  __jat__statcnt llog   MESSAGE.WARNING ;;
        pwarc)  __jat__statcnt P.llog MESSAGE.WARNING ;;
        rerrc)  __jat__statcnt llog   MESSAGE.ERROR   ;;
        perrc)  __jat__statcnt P.llog MESSAGE.ERROR   ;;
        rfailc) __jat__statcnt vlog   ASSERT.FAIL ;;
        pfailc) __jat__statcnt P.vlog ASSERT.FAIL ;;
        ppassc) __jat__statcnt P.vlog ASSERT.PASS ;;
        rpassc) __jat__statcnt vlog   ASSERT.PASS ;;
        pasrtc) __jat__statcnt P.vlog ASSERT.PASS ASSERT.FAIL ;;
        rasrtc) __jat__statcnt vlog   ASSERT.PASS ASSERT.FAIL ;;
        *) __jat__usage "invalid statistic field: $which"
           return 2 ;;
    esac
}

jat__meta() {
    #
    # Print fact $1 about session
    #
    # Usage:
    #    jat__meta bcaseid     # base CASEID (test id + phase CASEID, if any)
    #    jat__meta ptype       # current phase type
    #    jat__meta phint       # current phase hint
    #
    # Print meta-data about session.  Return with 1 if there's nothing
    # to print, eg. for `ptype` if we are outside phase.
    #
    local which=$1  # fact name
    local value
    case $which in
        bcaseid)  value=$(__jat__bcaseid) ;;
        ptype)    value=$(_jat_core__rd_keyr P.type | sed 's/^none$//') ;;
        phint)    value=$(_jat_core__rd_keyR P.hint) ;;
        *) __jat__usage "invalid meta-data field: $which"
           return 2 ;;
    esac
    test -n "$value" || return 1
    echo "$value"
}

jat__param() {
    #
    # Print value of test param name $1
    #
    # Print value of test parameter from $JAT__TEST_PARAMS.  Exit with
    # zero one if the value was non-empty, one if it was empty and two
    # if the value was not set at all.
    #
    # Note that *both* key and value have to be simple words.  See also
    # "PARAMETRIZATION" section of this manual.
    #
    local want=$1   # wanted parameter name
    local pair      # each pair in $JAT__TEST_PARAMS
    local pname     # each parameter value
    local pvalue    # each parameter name
    local es=0      # exit status
    for pair in ${JAT__TEST_PARAMS//,/ }; do
        case $pair in
            *=*)    pname=${pair%=*}; pvalue=${pair#"$pname="} ;;
            *)      jat__log_error "malformed parameter expression: $pair" ;;
        esac
        test "$pname" == "$want" || continue
        test -n "$pvalue"; es=$?
        echo "$pvalue"
        return $es
     done
     jat__log_error "no such parameter: $want in $JAT__TEST_PARAMS"
     return 2
}

jat__has_param() {
    #
    # True if test param name $1 is set
    #
    local want=$1   # wanted parameter name
    grep -qe ",$want=[^,=]*," <<<",$JAT__TEST_PARAMS,"
}

#shellcheck disable=SC2120
jat__pend() {
    #
    # End active phase
    #
    # Usage:
    #     jat__pend [-W]
    #
    # Close active phase and exit with status according to
    # the contents of the phase:
    #
    #     0 - if everything went fine
    #     1 - if there was a failed assert
    #     2 - if there was a warning message and -W was not specified
    #     4 - if there was error message
    #     5 - in case of other errors.
    #
    # Warnings can be suppressed from affecting exit status by
    # passing -W.   This allows to wrap a phase in a function and
    # use the function exit status to decide whether or not to abort
    # the test early, but still allow the function to raise flag in
    # in unclear circumstances.
    #
    local oldpdir   # current phase's workdir (if set by '-C')
    local oldpwd    # old workdir before jat__pstart?() changed it
    local duration  # phase duration
    local es=0      # exit status
    local es_w=true   # reflect warnings in exit status?
    while true; do case $1 in
        -W)     es_w=false; shift ;;
        "")     break ;;
        *)      __jat__usage "unknown argument: $1"
                return 5
                # ^^^ 2 has other meaning for this function
                ;;
    esac done
    oldpdir=$(_jat_core__rd_keyR P.pdir)
    oldpwd=$(_jat_core__rd_keyR P.PWD)
    test -n "$oldpdir" && {
        jat__log_info "changing back from phase dir: $oldpdir"
        popd >/dev/null \
         || jat__die "failed to go back from phase dir: $oldpdir to $oldpwd"
    }
    duration=$(__jat__toc jat__pstart)
    _jat_core__show_pend "$duration"
    __jat__log_event PHASE.END "$duration"
    test "$(jat__seen perr)"            && es=4
    $es_w && test "$(jat__seen pwar)"   && es=2
    test "$(jat__seen pfail)"           && es=1
    __jat__pdummy
    return $es
}

jat__pwrapd() {
    #
    # Start diagnostic phase, run command and end phase
    #
    # Usage:
    #      jat__pwrapd [phase-options] [--] CMD [ARGS]..
    #
    # Shorthand for
    #
    #     jat__pstartd [phase-options]
    #         CMD [ARGS]..
    #     jat__pend
    #
    # Exit status of the function is exit status of CMD, UNLESS something
    # (start/end of phase, or argument parsing) fails.
    #
    # See COMMON ARGUMENTS for details about phase options.
    #
    __jat__pwrap -t diag "$@"
}

jat__pwraps() {
    #
    # Start setup phase, run command and end phase
    #
    # Usage:
    #      jat__pwraps [phase-options] [--] CMD [ARGS]..
    #
    # Shorthand for
    #
    #     jat__pstarts [phase-options]
    #         CMD [ARGS]..
    #     jat__pend
    #
    # Exit status of the function is exit status of CMD, UNLESS something
    # (start/end of phase, or argument parsing) fails.
    #
    # See COMMON ARGUMENTS for details about phase options.
    #
    __jat__pwrap -t setup "$@"
}

jat__pwrapt() {
    #
    # Start testing phase, run command and end phase
    #
    # Usage:
    #      jat__pwrapt [phase-options] [--] CMD [ARGS]..
    #
    # Shorthand for
    #
    #     jat__pstartt [phase-options]
    #         CMD [ARGS]..
    #     jat__pend
    #
    # Exit status of the function is exit status of CMD, UNLESS something
    # (start/end of phase, or argument parsing) fails.
    #
    # See COMMON ARGUMENTS for details about phase options.
    #
    __jat__pwrap -t test "$@"
}

jat__pwrapc() {
    #
    # Start cleanup phase, run command and end phase
    #
    # Usage:
    #      jat__pwrapc [phase-options] [--] CMD [ARGS]..
    #
    # Shorthand for
    #
    #     jat__pstartc [phase-options]
    #         CMD [ARGS]..
    #     jat__pend
    #
    # Exit status of the function is exit status of CMD, UNLESS something
    # (start/end of phase, or argument parsing) fails.
    #
    # See COMMON ARGUMENTS for details about phase options.
    #
    __jat__pwrap -t cleanup "$@"
}

jat__pstartd() {
    #
    # Start diagnostic phase
    #
    # See COMMON ARGUMENTS section of this manual.
    #
    local es
    __jat__pstart -t diag "$@"; es=$?
    test $es -eq 2 && __jat__usage
    return $es
}

jat__pstartc() {
    #
    # Start cleanup phase
    #
    # See COMMON ARGUMENTS section of this manual.
    #
    local es
    __jat__pstart -t cleanup "$@"; es=$?
    test $es -eq 2 && __jat__usage
    return $es
}

jat__pstarts() {
    #
    # Start setup phase
    #
    # See COMMON ARGUMENTS section of this manual.
    #
    local es
    __jat__pstart -t setup "$@"; es=$?
    test $es -eq 2 && __jat__usage
    return $es
}

jat__pstartt() {
    #
    # Start test phase
    #
    # See COMMON ARGUMENTS section of this manual.
    #
    local es
    __jat__pstart -t test "$@"; es=$?
    test $es -eq 2 && __jat__usage
    return $es
}

jat__submit() {
    #
    # Submit file $1 as part of test result
    #
    # Usage:
    #
    #     jat__submit [-h HINT] [-H] [-s] [--] PATH [ALTNAME]
    #
    # Last element of PATH is used as result name, unless alternative
    # name ALTNAME is given, in which case it's used instead.
    #
    # HINT, if specified, will be used to describe relic in YAML
    # meta-data.  If HINT is not specified, the descripion will be
    # auto-generated from PATH, unless -H is provided, which
    # disables the auto-generation.
    #
    # By default, if submission fails, WARNING is logged. Use -s to
    # lower the log message severity to INFO.
    #
    local path              # PATH to submit
    local altname           # ALTNAME to submit file as
    local dest=$_JAT_CORE__RDIR/relics   # where to put submitted file
    local frdest            # final destination relative to $dest
    local nhint             # ALTNAME hint for message
    local hint              # relic description
    local ahint=true        # 'true' if hints should be auto-generated
    local onerr=jat__log_warning    # problem reporting function
    while true; do case $1 in
        -H) ahint=false; shift ;;
        -h) hint=$2; shift 2 || { __jat__usage "no HINT?"; return 2; } ;;
        -s) onerr=jat__log_info ; shift;;
        --) shift; break ;;
        *)  break ;;
    esac done
    path=$1; altname=$2
    test -n "$path" || {
        __jat__usage "no SRC?"
        return 2
    }
    mkdir -p "$dest" || {
        jat__log_error "cannot write to result storage: $dest"
        return 4
    }
    frdest="${path##*/}"
    test -n "$altname" && {
        dest="$dest/$altname"
        frdest="$altname"
        nhint=" as $altname"
        mkdir -p "$(dirname "$dest")"   # needed if altname has slashes
    }
    cp -ar -- "$path" "$dest" || {
        "$onerr" "result submission failed locally"
        return 3
    }
    test -z "$hint" && $ahint && test -f "$path" && hint="file: $path$nhint"
    test -z "$hint" && $ahint && test -d "$path" && hint="directory: $path$nhint"
    __jat__relic RELIC.FILE "$frdest" "$hint"
    _jath__relic RELIC.FILE "$(__jat__bcaseid)" "$frdest" "$hint" || :
}

jat__submitn() {
    #
    # Submit many files $@ as part of test result
    #
    # Usage:
    #
    #     jat__submitn [-s] [-h HINT] [-H] [-p PREFIX] PATH..
    #
    # Last element of PATH is used as result name, unless PREFIX
    # is also given, which is then prefixed to each final name.
    # Note that PREFIX can contain slashes, resulting in relative
    # paths; directories will be created as needed.
    #
    # HINT, if specified, will be used to describe relics in YAML
    # meta-data.  In this case, all relics will have the same
    # descripion.  If HINT is not specified, the descripion will be
    # auto-generated for each PATH, unless -H is provided, which
    # disables the auto-generation.
    #
    # By default, if any of submissions fail, WARNING is logged.
    # Use -s to lower the log message severity to INFO.
    #
    local path          # each PATH to submit
    local paths=()      # all PATHs to submit
    local prefix        # prefix to use in all PATHs
    local jsargs=()     # arguments for jat__submit
    local hint          # relic description
    local ahint=true    # 'true' if hints should be auto-generated
    local jscmd=()      # jat__submit command
    while true; do case $1 in
        -H) ahint=false; shift ;;
        -h) hint=$2; shift 2 || { __jat__usage "no HINT?"; return 2; } ;;
        -s) jsargs+=(-s);  shift ;;
        -p) prefix=$2; shift 2 || { __jat__usage "no PREFIX?"; return 2; } ;;
        --) shift; break ;;
        *)  break ;;
    esac done
    paths=("$@")
    debug -v prefix paths
    for path in "${paths[@]}"; do
        jscmd=(jat__submit)
        test -n "$hint" \
         && jscmd+=(-h "$hint")
        "$ahint" \
         || jscmd+=(-H)
        jscmd+=("$path")
        test -n "$prefix" \
         && jscmd+=("$prefix$(basename "$path")")
        debug -v jscmd
        "${jscmd[@]}"
    done
}

jat__submitb() {
    #
    # Submit tarball of files $2.. under name $1 as part of test result
    #
    # Usage:
    #
    #     jat__submitb [-s] [-h HINT] [-H] [-C CHDIR] [--] BALLNAME PATH..
    #
    # All PATHs are added to tarball, which is submitted under name
    # BALLNAME.  If CHDIR is provided, change to that directory
    # before packing; this makes it easier to control path prefix.
    #
    # HINT, if specified, will be used to describe the tarball in YAML
    # meta-data.  If HINT is not specified, the descripion will be
    # auto-generated for each PATH, unless -H is provided, which
    # disables the auto-generation.
    #
    # By default, if tarball creation fails, WARNING is logged.
    # Use -s to lower the log message severity to INFO.
    #
    local ballname      # submitted tarball name
    local opts=()       # extra tar options
    local tarball       # tarball path (in temp)
    local dest=$_JAT_CORE__RDIR/relics   # final storage path
    local frdest        # final destination relative to $dest
    local paths=()      # all PATHs to add to tarball
    local hint          # relic description
    local ahint=true    # 'true' if hints should be auto-generated
    local onerr=jat__log_warning    # problem reporting function
    while true; do case $1 in
        -H) ahint=false; shift ;;
        -h) hint=$2; shift 2 || { __jat__usage "no HINT?"; return 2; } ;;
        -s) onerr=jat__log_info; shift ;;
        -C) opts+=(-C "$2"); shift 2 || { __jat__usage "no CHDIR?"; return 2; } ;;
        --) shift; break ;;
        *)  break ;;
    esac done
    ballname=$1; shift
    paths=("$@")
    test -n "$ballname" || {
        __jat__usage "no BALLNAME?"
        return 2
    }
    test "${#paths[@]}" -gt 0 || {
        __jat__usage "no PATH?"
        return 2
    }
    mkdir -p "$dest" || {
        jat__log_error "cannot write to result storage: $dest"
        return 4
    }
    dest="$dest/$ballname"
    mkdir -p "$(dirname "$dest")"   # needed if ballname has slashes
    tarball=$(mktemp -t jat__submitb.tarball.XXXXXXXX)
    debug -v opts tarball paths
    tar -cz "${opts[@]}" -f "$tarball" "${paths[@]}" || {
        "$onerr" "failed to create tarball"
        rm "$tarball"
        return 4
    }
    frdest="$ballname"
    test -z "$hint" && $ahint && hint="tarball from: ${paths[*]}"
    __jat__relic RELIC.FILE "$frdest" "$hint"
    mv "$tarball" "$dest" || {
        warn "result submission failed locally"
        return 3
    }
    _jath__relic RELIC.FILE "$(__jat__bcaseid)" "$frdest" "$hint" || :
}

jat__submitc() {
    #
    # Submit stdout & stderr of code $1 as relic $2
    #
    # Usage:
    #
    #     jat__submitc [-h HINT] [-H] [-C CHDIR] [--] CODE RELNAME
    #
    # CODE is executed (if valid according to `bash -n`) and its
    # stdout is saved into file named RELNAME and submitted as
    # part of the test result.
    #
    # HINT, if specified, will be used to describe the relic in YAML
    # meta-data.  If HINT is not specified, the descripion will be
    # auto-generated, unless -H is provided, which disables the
    # auto-generation.
    #
    # By default, CODE is executed in current directory.  Use
    # -C to specify alternate directory just for the code execution.
    #
    local code          # CODE to run
    local relname       # relic file name
    local dest=$_JAT_CORE__RDIR/relics   # final storage path
    local frdest        # final destination relative to $dest
    local hint          # relic description
    local ahint=true    # 'true' if hints should be auto-generated
    local wdir=.        # working directory for the CODE
    while true; do case $1 in
        -C) wdir=$2; shift 2 || { __jat__usage "no CHDIR?"; return 2; } ;;
        -H) ahint=false; shift ;;
        -h) hint=$2; shift 2 || { __jat__usage "no HINT?"; return 2; } ;;
        --) shift; break ;;
        *)  break ;;
    esac done
    code=$1
    relname=$2
    test -n "$code" || {
        __jat__usage "no CODE?"
        return 2
    }
    bash -n <<<"$code" || {
        jat__log_error "syntax error in CODE: $code"
        return 2
    }
    test -n "$relname" || {
        __jat__usage "no RELNAME?"
        return 2
    }
    mkdir -p "$dest" || {
        jat__log_error "cannot write to result storage: $dest"
        return 3
    }
    dest="$dest/$relname"
    mkdir -p "$(dirname "$dest")"   # needed if relname has slashes
    debug -v code relname path
    (
        cd "$wdir" || {
            jat__log_error "could not change directory to: $wdir"
            return 3
        }
        eval "$code"
    ) >"$dest"
    frdest="$relname"
    test -z "$hint" && $ahint && hint="output of: \`$code\`"
    __jat__relic RELIC.FILE "$frdest" "$hint"
    _jath__relic RELIC.FILE "$(__jat__bcaseid)" "$frdest" "$hint" || :
}

__jat__assert() {
    #
    # Make assert event
    #
    # Use this function to create own asserts
    #
    case $1 in
        ASSERT.PASS)   _jat_core__show_pass "$2" ;;
        ASSERT.FAIL)   _jat_core__show_fail "$2" ;;
    esac
    __jat__log_event "$@"
}

__jat__relic() {
    #
    # Make relic event
    #
    # Use this function to attach independent relic data to log
    #
    case $1 in
        RELIC.DATA) _jat_core__show_rdata "$@" ;;
        RELIC.FILE) _jat_core__show_rfile "$@" ;;
    esac
    __jat__log_event "$@"
}

__jat__bcaseid() {
    #
    # Print base CASEID (test id + phase CASEID if any)
    #
    local pcaseid
    pcaseid=$(_jat_core__rd_keyR P.pcaseid)
    echo -n "$JAT__TEST_ID"
    test -n "$JAT__TEST_PARAMS" && echo -n "@$(__jat__normalize_params)"
    test -n "$pcaseid" && echo -n ":$pcaseid"
    return 0
}

__jat__cmp_match() {
    #
    # True if $RVal matches $OVal by $Op
    #
    case $Op in
        eq|ne|lt|gt|le|ge) test "$RVal" -"$Op" "$OVal" ;;
        ==)                test "$RVal" == "$OVal" ;;
        re)                grep -qxe "$OVal" <<< "$RVal" ;;
        *)  return 2 ;;
    esac
}

__jat__es_match() {
    #
    # True if exit status $2 matches expression $1
    #
    local expr=$1   # ES_EXPR expression
    local es=$2     # compared exit status
    local part      # each part of ES_EXPR
    for part in ${expr//,/ }; do
        test -n "$part" || continue
        #FIXME: a rather funny implementation (works, though...)
        eval "echo {${part/-/..}}" | grep -qwFe "$es" && return 0
    done
    return 1
}

__jat__final_es() {
    #
    # Print final exit status
    #
    # Roughly follows TFKit's semantic; unoffical link:
    #
    #     https://gitlab.com/vornet/infra/tfkit/-/tree/master/src/doc
    #
    test "$(jat__stat rerrc)"  -gt 0 && echo 4 && return
    test "$(jat__stat rwarc)"  -gt 0 && echo 2 && return
    test "$(jat__stat rfailc)" -gt 0 && echo 1 && return
    echo 0
}

__jat__final_result() {
    #
    # Print final exit status
    #
    # Roughly follows TFKit's semantic; unoffical link:
    #
    #     https://gitlab.com/vornet/infra/tfkit/-/tree/master/src/doc
    #
    test "$(jat__stat rerrc)"  -gt 0 && echo error && return
    test "$(jat__stat rwarc)"  -gt 0 && echo warning && return
    test "$(jat__stat rfailc)" -gt 0 && echo fail && return
    echo pass
}

__jat__log_event() {
    #
    # Log event to YAML log via __jat__writelog()
    #
    # Usage:
    #
    #     __jat__log_event RUN.START|RUN.RELOAD|RUN.END
    #     __jat__log_event PHASE.END DURATION
    #     __jat__log_event PHASE.START ID HINT TYPE [K=V]
    #     __jat__log_event ASSERT.PASS|ASSERT.FAIL HINT CASEID [K=V]..
    #     __jat__log_event MESSAGE.INFO|MESSAGE.WARNING|MESSAGE.ERROR MSG [K=V]
    #     __jat__log_event RELIC.DATA K=V..
    #     __jat__log_event RELIC.FILE PATH
    #
    local EType=$1; shift           # event type
    local meatfn                    # event module function
    local origin=${FUNCNAME[1]}     # rvent origin (calling function)
    local BaseCaseId                # CASEID base (jats://...)
    BaseCaseId=$(__jat__bcaseid)
    case $EType in
        ASSERT.PASS|ASSERT.FAIL) : ;;
        MESSAGE.INFO|MESSAGE.ERROR|MESSAGE.WARNING) : ;;
        PHASE.START|PHASE.END) : ;;
        RUN.START|RUN.RELOAD|RUN.END) : ;;
        RELIC.DATA|RELIC.FILE) : ;;
        *)  __jat__usage "bad ETYPE: $EType"
            return 2 ;;
    esac
    case $EType:$origin in
        PHASE.END:jat__pend)                meatfn="__jat__log_evt_pend" ;;
        PHASE.START:__jat__pstart)          meatfn="__jat__log_evt_pstart" ;;
        RUN.END:__jat__rfinish)             meatfn="true" ;;
        RUN.RELOAD:jat__rinit)              meatfn="true" ;;
        RUN.START:jat__rinit)               meatfn="true" ;;
        ASSERT.FAIL:__jat__assert)          meatfn="__jat__log_evt_assert" ;;
        ASSERT.PASS:__jat__assert)          meatfn="__jat__log_evt_assert" ;;
        MESSAGE.ERROR:__jat__usage)         meatfn="__jat__log_evt_message" ;;
        MESSAGE.ERROR:__jat__log)           meatfn="__jat__log_evt_message" ;;
        MESSAGE.ERROR:__jat__logv)          meatfn="__jat__log_evt_message" ;;
        MESSAGE.INFO:__jat__log)            meatfn="__jat__log_evt_message" ;;
        MESSAGE.INFO:__jat__logv)           meatfn="__jat__log_evt_message" ;;
        MESSAGE.WARNING:__jat__log)         meatfn="__jat__log_evt_message" ;;
        MESSAGE.WARNING:__jat__logv)        meatfn="__jat__log_evt_message" ;;
        RELIC.DATA:__jat__relic)            meatfn="__jat__log_evt_rdata" ;;
        RELIC.FILE:__jat__relic)            meatfn="__jat__log_evt_rfile" ;;
        *)  _jat_core__show_error "illegal call of __jat__log_event: $EType from $origin"
            EType=MESSAGE.ERROR ;;
    esac
    echo "$EType" >> "$(_jat_core__rd_path "vlog")"
    echo "$EType" >> "$(_jat_core__rd_path "P.vlog")"
    {
        echo "    -"
        echo "        etype: $EType"
        echo "        stamp: $(__jat__newstamp)"
        $meatfn "$@"
    } | __jat__writelog
    __jat__hrep_evt "$@"
}

__jat__log() {
    #
    # Log message with severity $1 and lines $2..
    #
    # Usage:
    #     __jat__log SEVERITY [LINE]..
    #
    # Create message (MESSAGE.* event in YAML log) consisting of provided
    # LINEs.
    #
    local sev=${1,,}; shift   # severity
    local line          # each line of message
    local msg           # complete message
    local head=true     # true if joining first item
    local hdline        # hexdump line
    local showfn="_jat_core__show_t$sev"
    type -t "$showfn" >/dev/null || {
        warn "bug: invalid severity (not 'info|error|warning'): '$sev'"
        return 2
    }
    for line in "$@"; do
        if test "$(__jat__qtype "$line")" == x; then
            "$showfn" "(binary detected; forcing hexdump mode)"
            hexdump -C <<<"$line" \
              | while read -r hdline; do
                   "$showfn" "$hdline"
                done
        else
            "$showfn" "$line"
        fi
        $head || msg+=$'\n'; head=false
        msg+="$line"
    done
    echo "MESSAGE.${sev^^}" >> "$(_jat_core__rd_path "llog")"
    echo "MESSAGE.${sev^^}" >> "$(_jat_core__rd_path "P.llog")"
    __jat__log_event "MESSAGE.${sev^^}" "$msg"
}

__jat__logv() {
    #
    # Log contents of internal variables named $2.. with severity $1
    #
    # Usage:
    #     __jat__logv SEVERITY VARNAME [VARNAME..]
    #
    # Creates log message showing current values of all given variables,
    # or string `#Unset` for undefined variables.
    #
    # Every VARNAME must be a valid variable name according to rules
    # described as "simple words" in STRINGS section of this manual.
    #
    local __jat__sev=${1,,}; shift   # severity
    local __jat__varname       # variable name
    local __jat__varbody       # variable value representation text
    local __jat__msg           # message
    local __jat__head=true     # true if joining first item
    local __jat__showfn="_jat_core__show_t$__jat__sev"
    type -t "$__jat__showfn" >/dev/null || {
        warn "bug: invalid severity (not 'info|error|warning'): '$__jat__sev'"
        return 2
    }
    for __jat__varname in "$@"; do
        isa__name "$__jat__varname" || {
            __jat__usage "unsafe __jat__varname: $__jat__varname"
            return 2
        }
        if declare_str=$(declare -p "$__jat__varname" 2>/dev/null); then
            __jat__varbody="${declare_str#declare ?? }"
        else
            __jat__varbody="$__jat__varname #Unset"
        fi
        "$__jat__showfn" "$__jat__varbody"
        $__jat__head || __jat__msg+=$'\n'; __jat__head=false
        __jat__msg+="$__jat__varbody"
    done
    echo "MESSAGE.${__jat__sev^^}" >> "$(_jat_core__rd_path "llog")"
    echo "MESSAGE.${__jat__sev^^}" >> "$(_jat_core__rd_path "P.llog")"
    __jat__log_event "MESSAGE.${__jat__sev^^}" "$__jat__msg"
}

__jat__normalize_params() {
    #
    # Print all pairs from $JAT__TEST_PARAMS sorted
    #
    local pair      # each pair in $JAT__TEST_PARAMS
    for pair in ${JAT__TEST_PARAMS//,/ }; do
        echo "$pair"
    done \
      | LC_ALL=C sort \
      | paste -sd,
}

__jat__hrep_evt() {
    #
    # Report event to harness
    #
    case $EType in
        PHASE.*)
            _jath__phase \
                "$EType" \
                "$BaseCaseId" \
                "$(_jat_core__rd_keyr P.hint)" \
                "$(_jat_core__rd_keyr P.type)" \
                "$(_jat_core__presult)"
            ;;
        MESSAGE.*)
            _jath__message \
                "$EType" \
                "$BaseCaseId" \
                "$1"
            ;;
        RUN.*)
            _jath__run \
                "$EType" \
                "$BaseCaseId" \
                "$(__jat__final_result)"
            ;;
    esac
}

__jat__log_evt_pend() {
    #
    # Event module for PHASE.END
    #
    # Usage:
    #
    #     __jat__log_event PHASE.END DURATION
    #
    local duration=$1; shift    # phase duration
    echo "        phase:"
    echo "            duration: $duration"
}

__jat__log_evt_pstart() {
    #
    # Event module for PHASE.START
    #
    # Usage:
    #
    #     __jat__log_event PHASE.START ID HINT TYPE [K=V]
    #
    local pid=$1; shift     # phase ID
    local phint=$1; shift   # phase HINT
    local ptype=$1; shift   # phase TYPE
    local data=("$@")       # phase parameters
    echo "        phase:"
    echo "            id: $pid"
    echo "            type: $ptype"
    echo "            hint: |-4"
    echo "                $phint"
    __jat__yamld 2 data "${data[@]}"
}

__jat__log_evt_message() {
    #
    # Event module for MESSAGE.*
    #
    # Usage:
    #
    #     __jat__log_event MESSAGE.INFO MSG [K=V]..
    #     __jat__log_event MESSAGE.WARNING MSG [K=V]..
    #     __jat__log_event MESSAGE.ERROR MSG [K=V]..
    #
    local msg=$1; shift     # message body
    local data=("$@")       # arbitrary data
    __jat__yamls 2 message "$msg"
    __jat__yamld 2 data "${data[@]}"
}

__jat__log_evt_rdata() {
    #
    # Event module for RELIC.DATA
    #
    # Usage:
    #
    #     __jat__log_event RELIC.DATA K=V..
    #
    local data=("$@")       # arbitrary data
    __jat__yamld 2 data "${data[@]}"
}

__jat__log_evt_rfile() {
    #
    # Event module for RELIC.FILE
    #
    # Usage:
    #
    #     __jat__log_event RELIC.FILE PATH [HINT]
    #
    local path=$1
    local hint=$2
    local data=("path=$path")
    test -n "$hint" && data+=("hint=$hint")
    __jat__yamld 2 data "${data[@]}"
}

__jat__log_evt_assert() {
    #
    # Event module for ASSERT.*
    #
    # Usage:
    #
    #     __jat__log_event ASSERT.PASS HINT CASEID [K=V]..
    #     __jat__log_event ASSERT.FAIL HINT CASEID [K=V]..
    #
    local hint=$1; shift    # assert HINT
    local caseid=$1; shift  # assert CASEID
    local data=("$@")       # assert data (K=V pairs)
    case $caseid in
        "") caseid=$BaseCaseId ;;
        *)  caseid=$BaseCaseId:$caseid ;;
    esac
    __jat__yamls 2 hint "$hint"
    __jat__yamls 2 caseid "$caseid"
    __jat__yamld 2 data "${data[@]}"
}

__jat__newstamp() {
    #
    # Create new timestamp
    #
    case $__JAT__DETERMINISTIC in
        true)   _jat_core__bumpid time ;;
        false)  date +%s ;;
        *) die "bad value of __JAT__DETERMINISTIC: $__JAT__DETERMINISTIC" ;;
    esac
}

__jat__newsid() {
    #
    # Create new timestamp
    #
    case $__JAT__DETERMINISTIC in
        true)   echo noid ;;
        false)  date +%s-%N ;;
        *) die "bad value of __JAT__DETERMINISTIC: $__JAT__DETERMINISTIC" ;;
    esac
}

__jat__pdummy() {
    #
    # Create dummy phase
    #
    _jat_core__rd_keyw phase "dummy"
    _jat_core__rd_keyw P.type "none"
}

__jat__pwrap() {
    #
    # Start diagnostic phase, run command and end phase
    #
    # Usage:
    #      __jat__pwrap -t TYPE [phase-options] [--] CMD [ARGS]..
    #
    # Shorthand for
    #
    #     __jat__pstart -t TYPE [phase-options]
    #         CMD [ARGS]..
    #     jat__pend
    #
    # See COMMON ARGUMENTS for details about phase options.
    #
    local ptype
    local pargs=()
    local cmd=()
    local es
    while true; do case $1 in
        -t)       ptype=$2;           shift 2 || { __jat__usage; return 2; } ;;
        -C|-c|-h) pargs+=("$1" "$2"); shift 2 || { __jat__usage; return 2; } ;;
        --)       shift; break ;;
        -*)       __jat__usage; return 2 ;;
        *)        break ;;
    esac done
    cmd=("$@")
    debug -v ptype pargs cmd
    __jat__pstart -t "$ptype" "${pargs[@]}" || return $?
        "${cmd[@]}"; es=$?
    #shellcheck disable=SC2119
    jat__pend || return $?
    return $es
}

__jat__pstart() {
    #
    # Start phase
    #
    local type      # phase type
    local hint      # ^^ hint
    local pdir      # ^^ directory
    local pcaseid   # ^^ case id, if assigned
    local oldphase  # current phase id
    local newphase  # new phase id
    while true; do case $1 in
        -h) hint="$2"; shift 2 || return 2 ;;
        -t) type="$2"; shift 2 || return 2 ;;
        -c) pcaseid="$2"; shift 2 || return 2 ;;
        -C) pdir="$2"; shift 2 || return 2 ;;
        *)  break ;;
    esac done
    test $# -gt 0 && return 2
    oldphase=$(_jat_core__rd_keyr phase)
    test "$oldphase" == dummy || {
        jat__log_error "old phase not ended; ending automatically: $oldphase"
        #shellcheck disable=SC2119
        jat__pend
    }
    debug -v type hint pdir
    newphase=$(_jat_core__bumpid phasid)
    _jat_core__rd_keyw phase "$newphase"
    _jat_core__rd_keyw P.type "$type"
    _jat_core__rd_keyw P.hint "$hint"
    _jat_core__rd_keyw P.pcaseid "$pcaseid"
    _jat_core__show_pstart
    __jat__log_event PHASE.START "$newphase" "$hint" "$type" \
        "pcaseid=$pcaseid" "pdir=$pdir"
    __jat__tic jat__pstart
    test -n "$pdir" && {
        _jat_core__rd_keyw P.pdir "$pdir"
        _jat_core__rd_keyw P.PWD "$PWD"
        jat__log_info "changing directory for phase: $pdir"
        pushd "$pdir" >/dev/null || {
            jat__log_error "failed to change to phase directory: $pdir"
            return 4
        }
    }
    return 0
}

__jat__statcnt() {
    #
    # Sum up number of lines $2.. in run data key $1
    #
    # Print count of lines or 0 if the key is not present at all.
    #
    local key=$1; shift     # run data key
    local line              # each line to find
    local rdpath            # run data path
    local sum=0             # sum of all lines' occurrences
    local lsum=0            # each line occurrence number
    rdpath=$(_jat_core__rd_path "$key")
    test -f "$rdpath" || { echo 0; return 1; }
    test -s "$rdpath" || { echo 0; return 1; }
    for line in "$@"; do
        lsum=$(grep -cxFe "$line" "$rdpath")
        sum=$((sum + lsum))
    done
    echo "$sum"
    test "$sum" -gt 0
}

__jat__tic() {
    #
    # Make reference timestamp named $1
    #
    local key=$1
    local now
    now=$(date +%s)
    _jat_core__rd_keyw "tictoc/$key" "$now"
}

__jat__toc() {
    #
    # Print duration since reference timestamp $1
    #
    local key=$1
    local ref
    local now
    $__JAT__DETERMINISTIC && echo 0 && return 0
    now=$(date +%s)
    ref=$(_jat_core__rd_keyr "tictoc/$key")
    echo "$((now - ref))"
}

__jat__usage() {
    #
    # Print usage error $1 and hint according to FUNCNAME
    #
    local msg=$1    # usage message
    local parent=${FUNCNAME[1]}   # parent function (assuming usage thereof)
    local patt      # each pattern to mentor on
    local patts=()  # acceptable usage patterns
    local lines=()  # error lines
    lines=("bad usage: $parent(): $msg")
    case $parent in
        jat__abort)    patts=("REASON [--] [LINE].." "REASON -f [FILE].." "REASON -c CMD [ARG].." "") ;;
        jat__cmd)      patts=("[assert-options] [-o R_OUT] [-e R_ERR] [-s R_ESF] [-S ES_EXPR] [-V ES_EXPR] [--] CMD [ARG]..") ;;
        jat__eval)     patts=("[assert-options] [-o R_OUT] [-e R_ERR] [-s R_ESF] [-S ES_EXPR] [-V ES_EXPR] [--] CODE") ;;
        jat__cmp)      patts=("[assert-options] RVAL OP OVAL") ;;
        jat__fail)     patts=("[assert-options]") ;;
        jat__log_error)     patts=("[LINE]..") ;;
        jat__log_errorv)    patts=("[VARNAME]..") ;;
        jat__log_info)      patts=("[LINE]..") ;;
        jat__log_infov)     patts=("[VARNAME]..") ;;
        jat__log_warning)   patts=("[LINE]..") ;;
        jat__log_warningv)  patts=("[VARNAME]..") ;;
        jat__meta)     patts=(bcaseid ptype) ;;
        jat__pass)     patts=("[assert-options]") ;;
        jat__pend)     patts=("[-W]") ;;
        jat__pstartc)  patts=("[phase-options]") ;;
        jat__pstartd)  patts=("[phase-options]") ;;
        jat__pstarts)  patts=("[phase-options]") ;;
        jat__pstartt)  patts=("[phase-options]") ;;
        jat__rfinish)  patts=("") ;;
        jat__rinit)    patts=("") ;;
        jat__submit)   patts=("[-s] [-h HINT] [-H] SRC [ALTNAME]") ;;
        jat__submitb)  patts=("[-s] [-h HINT] [-H] [-C CHDIR] BALLNAME PATH..") ;;
        jat__submitc)  patts=("[-h HINT] [-H] [-C CHDIR] CMD [ARG]..") ;;
        jat__submitn)  patts=("[-s] [-h HINT] [-H] [-p PREFIX] PATH..") ;;
        jat__seen)     patts=(rfail pfail rerr ..) ;;
        jat__stat)     patts=(rpassc rfailc ppassc pfailc ..) ;;
    esac
    for patt in "${patts[@]}"; do
        lines+=("  usage: $parent $patt")
    done
    jat__log_error "${lines[@]}"
}

__jat__writelog() {
    #
    # Write logged output (copy if needed)
    #
    __jat__strip4yaml \
      | case $JAT__YLOG in
            "") cat >> "$_JAT_CORE__RDIR/log.yaml" ;;
            -)  tee -a "$_JAT_CORE__RDIR/log.yaml" ;;
            *)  tee -a "$_JAT_CORE__RDIR/log.yaml" >> "$JAT__YLOG" ;;
        esac
}

__jat__valid_pair() {
    #
    # True if $1 is a valid K=V pair
    #
    local tainted=$1    # checked data
    grep -qx '\([tor][.]\)\?[[:alpha:]_][[:alnum:]_]*=.*' <<<"$tainted"
}

__jat__strip4yaml() {
    #
    # Replace yaml-invalid characters
    #
    perl -CS -Mutf8 -pe "tr/$__JAT__YAML_BAD/$__JAT__YAML_REPL/"
}

__jat__yamlqs() {
    #
    # Put stdin in a quoted scalar string indented by $1
    #
    # Use backslash in basic cases and rest quote using \U notation.
    #
    local sp=$1                     # space prefix
    perl -CS -Mutf8 -ne '
        sub escln {
            my $in = shift;
            my $out;
            my $chr;
            for $chr (split //, $in) {
                if ($chr =~ m|\N{U+0}|) {
                    $out .= "\\0";
                } elsif ($chr =~ m|\N{U+7}|) {
                    $out .= "\\a";
                } elsif ($chr =~ m|\N{U+8}|) {
                    $out .= "\\b";
                } elsif ($chr =~ m|\N{U+9}|) {
                    $out .= "\\t";
                } elsif ($chr =~ m|\N{U+B}|) {
                    $out .= "\\v";
                } elsif ($chr =~ m|\N{U+C}|) {
                    $out .= "\\f";
                } elsif ($chr =~ m|\N{U+D}|) {
                    $out .= "\\r";
                } elsif ($chr =~ m|\N{U+5C}|) {
                    $out .= "\\";
                } elsif ($chr =~ m|['"$__JAT__YAML_BAD"']|) {
                    my $u = ord($chr);
                    $out .= sprintf("\\u%x", $u);
                } else {
                    $out .= $chr;
                }
            }
            return "'"$sp"'    $out";
        }
        print escln($_);
    '
}

__jat__yamls() {
    #
    # Print a scalar field named $2 with value $3, nested $1 levels
    #
    local nest=$1; shift    # nesting level
    local name=$1           # field name
    local value=$2          # scalar value
    local qtype=$3          # quoting type
    local sp                # indentation space
    sp=$(__jat__yamlnest "$nest")
    case $value in
        "") echo "$sp$name: ~" ;;
        *)
            test -n "$qtype" || qtype=$(__jat__qtype "$value")
            case $qtype in
                p)
                    echo "$sp$name: |-4"
                    sed "s/^/$sp    /"
                    ;;
                q)
                    echo "$sp$name+$qtype: |-4"
                    __jat__yamlqs "$sp"
                    ;;
                x)
                    echo "$sp$name+$qtype: |-4"
                    hexdump -C \
                      | sed "s/^/$sp    /"
                    ;;
                *)
                    warn "invalid quoting type: $qtype"
                    return 3
                    ;;
            esac <<<"$value"
            ;;
    esac
}

__jat__yamla() {
    #
    # Print an array field named $2 with array of values $3.., nested $1 levels
    #
    local nest=$1; shift    # nesting level
    local name=$1; shift    # field name
    local value             # each value in array
    local sp                # indentation space
    sp=$(__jat__yamlnest "$nest")
    test $# -eq 0 && echo "    $name: []" && return 0
    echo "$sp$name:"
    for value in "$@"; do
        case $value in
            "") echo "$sp    - ~" ;;
            *)  echo "$sp    - |-4"
                echo "$sp        $value" ;;
        esac
    done
}

__jat__yamlnest() {
    #
    # Print $1 x 4 of spaces
    #
    local n=$1
    while test "$n" -gt 0; do
        echo -n '    '
        (( n-- ))
    done
}

__jat__yamld() {
    #
    # Print a dict field named $2 and composed of KEY=VALUE pairs $3.., nested $1 levels
    #
    local nest=$1; shift    # nesting level
    local name=$1; shift    # field name
    local pair              # each KEY=VALUE pair
    local key               # each key in dict
    local value             # each value in dict
    local qtype             # quoting type
    local sp                # indentation space
    sp=$(__jat__yamlnest "$nest")
    test $# -eq 0 && echo "$sp$name: {}" && return 0
    echo "$sp$name:"
    for pair in "$@"; do
        key=${pair%%=*}; value=${pair#"$key="}
        case $value in
            "") echo "$sp    $key: ~" ;;
            *)
                qtype=$(__jat__qtype "$value")
                case $qtype in
                    p)
                        echo "$sp    $key: |-4"
                        sed "s/^/$sp        /"
                        ;;
                    q)
                        echo "$sp    $key+$qtype: |-4"
                        __jat__yamlqs "$sp    "
                        ;;
                    x)
                        echo "$sp    $key+$qtype: |-4"
                        hexdump -C \
                          | sed "s/^/$sp        /"
                        ;;
                    *)
                        warn "invalid quoting type: $qtype"
                        return 3
                        ;;
                esac <<<"$value"
                ;;
        esac
    done
}

__jat__qtype() {
    #
    # Select quoting type for data $1
    #
    # Detect one of three options:
    #
    #  *  'p' if all is plain UTF-8 text that can be printed right into
    #     the YAML file.  Note that this excludes control chars and
    #     some other other "dangerous" characters that could cause
    #     problems further down the line.
    #
    #     This means the data should be left alone.
    #
    #  *  'q' for everything else, if it's still recognizable as UTF-8.
    #
    #     This allows including characters from previous category, by
    #     using Python-like backslash quoting sequences (\a, .. \t,
    #     \uNNN..)
    #
    #  *  'x' for just binary data.
    #
    #     This means the data should be hexdumped.
    #
    local value=$1      # value to check
    local needq         # range that needs quoting
    local qt            # decided quoting type
    needq+='\N{U+0}-\N{U+9}'
    needq+='\N{U+B}-\N{U+1F}'
    needq+='\N{U+7F}'
    needq+=$__JAT__YAML_BAD
    qt=$(
        perl -w -C0 -Mutf8 -ne 'utf8::decode($_) or exit(1)' \
            <<<"$value" 2>/dev/null \
         || echo x
    )
    test -n "$qt" && echo "$qt" && return
    qt=$(
        perl -CS -Mutf8 -ne "
            if (m|[$needq]|) {
                print('q');
                exit(0);
            }
        " <<<"$value"
    )
    test -n "$qt" && echo "$qt" && return
    echo p
}

__shellfu_jat__init() {
    #
    # Initialize 'jat' module
    #
    if test -z "$__JAT__YAML_BAD"; then
        #
        # Yaml won't allow all characters:
        #
        # > [...] The allowed character range explicitly excludes the C0 control
        # > block #x0-#x1F (except for TAB #x9, LF #xA, and CR #xD which are
        # > allowed), DEL #x7F, the C1 control block #x80-#x9F (except for NEL
        # > #x85 which is allowed), the surrogate block #xD800-#xDFFF, #xFFFE,
        # > and #xFFFF.
        #
        # Due to bug in some versions of PyYAML, also replace several extra
        # chars; that is, everything above U+10000.
        #
        __JAT__YAML_BAD+='\N{U+0}-\N{U+8}\N{U+B}\N{U+C}\N{U+E}-\N{U+1F}'    # C0 with some gaps
        __JAT__YAML_BAD+='\N{U+7F}'                                         # DEL alone
        __JAT__YAML_BAD+='\N{U+80}-\N{U+84}\N{U+86}-\N{U+9F}'               # C1 with NEL gap
        # __JAT__YAML_BAD+='\N{U+D800}-\N{U+DFFF}'                          # surrogates
        # __JAT__YAML_BAD+='\N{U+FFFE}-\N{U+FFFF}'                          # 0xFFFE and 0xFFFF
        #FIXME: for some reasons perl complains about these ^^
        __JAT__YAML_BAD+='\N{U+10000}-\N{U+FFFFFF}'               # because of buggy parser(s)
        #FIXME: remove after https://bugzilla.redhat.com/show_bug.cgi?id=1677421
    fi
}

#shellfu module-version=0.2.12
