#!/bin/bash

shellfu import arr
shellfu import pretty
shellfu import sfpi

shellfu import jat
shellfu import _jat_core

#
# JAT context utilities
#
# JAT contexts are a mechanism for maintaining common test setups that
# can be re-used across test suite.
#
# Context consists of one or more combinations of classes and their
# parameters, where for each class, a plug-in must exist that implements
# its setup and cleanup funcions.
#
#
# CONTEXTS
# --------
#
# A context is created by call to jat_ctx__setup, with set of context
# parameters.   Context parameters are prefixed with class name, which
# correspond to context plugin implementing the given parameter.
#
# For example, setup call like this:
#
#     jat_ctx__setup \
#         foo.@time=day \
#         foo.bar=baz \
#         foo.bar=baz2 \
#         foo.quux=7 \
#         abc.a=1
#
# will create context with 5 parameters, 4 of them implemented of 'foo' class
# implemented by 'foo' plugin and 1 of them of 'abc' class, i.e. 'abc' plugin.
#
# During setup, are satisfied separately in arbitrary order.  Every parameter
# results in own call of the respective plugin handler, with exception of
# parameters whose names start with '@' (at sign); these don't warrant extra
# plugin calls.
#
# Contexts can be "stacked", so calling jat_ctx__setup() three times will
# instantiate three contexts.  Contexts are tore down by jat_ctx__cleanup
# function in the opposite order than they were instantiated.
#
#
# CONTEXT PLUGINS
# ---------------
#
# Context plugins are Shellfu modules named in following manner:
#
#     _jat_ctxp_CLASS.sh
#
# where '_jat_ctxp_' is a constant prefix, and CLASS is name of the
# context class this plugin implements.  Each of this plugins must
# contain at least these functions, so-called handlers:
#
#     _jat_ctxp_CLASS__setup()
#         :
#     }
#
#     _jat_ctxp_CLASS__cleanup()
#         :
#     }
#
# where cleanup handler must undo all operations of setup handler.
# Both handlers are guarranteed to be executed in the same directory
# dedicated for given context.  That is, cleanup handler can trivially
# access files leftover by the respective setup handler.
#
# Optionally a verify handler can be implemented
#
#     _jat_ctxp_CLASS__verify()
#         :
#     }
#
# When called, handler gets passed two arguments: parameter name and
# parameter value.  For example, setup of context like this:
#
#     jat_ctx__setup \
#         foo.bar=baz \
#         foo.bar=baz2 \
#         foo.quux=7 \
#         abc.a=1
#
# will result in following calls:
#
#     _jat_ctxp_foo__setup bar baz
#     _jat_ctxp_foo__setup bar baz2
#     _jat_ctxp_foo__setup quux 7
#     _jat_ctxp_abc__setup a 1
#
# If handlers also defined verify handlers, they will be also called:
#
#     _jat_ctxp_foo__verify bar baz
#     _jat_ctxp_abc__verify a 1
#     _jat_ctxp_foo__verify bar baz2
#     _jat_ctxp_foo__verify quux 7
#
# Note that the order of calls is arbitrary and has nothing to do
# with order of parameter appearance on the jat_ctx__setup command
# line.  What's guarranteed is that within a context setup,  all
# verify handlers are called after all setup handlers.
#
# To make use of parameters passed on setup, both setup and cleanup
# handlers have access to function _jat_ctx__param().  This function
# accepts single argument of parameter name, and prints value of that
# parameter.
#

jat_ctx__cleanup() {
    #
    # Clean up all or last set up context
    #
    # Usage:
    #     jat_ctx__cleanup [-a]
    #
    # Clean up last recently applied context.
    #
    # If `-a` is passed, clean up all contexts in reverse order than the one
    # in which they have been set up.
    #
    # See CONTEXT PLUGINS for details on cleanup handlers.
    #
    local ctxlog
    local golxtc
    local clslog
    local golslc
    local __JAT_CTX__FN=jat_ctx__cleanup
    local __JAT_CTX__ID
    local single=true
    local usage="usage: jat_ctx__cleanup [-a]"
    while true; do case $1 in
        -a) single=false; shift ;;
        "") break ;;
        *)  jat__log_error "$usage"; return 2 ;;
    esac done
    test -n "$1" && { jat__log_error "$usage"; return 2; }
    ctxlog=$(_jat_core__rd_path "contexts/log")
    test -s "$ctxlog" || {
        jat__log_error "no contexts to cleanup"
        return 3
    }
    golxtc=$(_jat_core__rd_path "contexts/gol")
    tac "$ctxlog" >"$golxtc"
    while test -s "$golxtc"; do
        read -r __JAT_CTX__ID <"$golxtc"
        test "$__JAT_CTX__ID" == "$(tail -1 "$ctxlog")"  \
         || jat__log_error "internal context tracking bug"
        sed -i '1d' "$golxtc"; sed -i '$d' "$ctxlog"
        clslog=$(_jat_core__rd_path "contexts/$__JAT_CTX__ID/log")
        golslc=$(_jat_core__rd_path "contexts/$__JAT_CTX__ID/gol")
        tac "$clslog" >"$golslc"
        while test -s "$golslc"; do
            IFS='=' read -r __JAT_CTX__CLASS <"$golslc"
            test "$__JAT_CTX__CLASS" == "$(tail -1 "$clslog")"  \
             || jat__log_error "internal class tracking bug"
            sed -i "/$__JAT_CTX__CLASS/d" "$golslc"; sed -i "/$__JAT_CTX__CLASS/d" "$clslog"
            jat__log_info "$__JAT_CTX__FN():$__JAT_CTX__ID: cleaning up: $__JAT_CTX__CLASS"
            __jat_ctx__call cleanup || {
                jat__log_error "context cleanup handler reported error"
                return 4
            }
        done
        $single && break
    done <<<"$(tac "$ctxlog")"
    return 0
}

jat_ctx__setup() {
    #
    # Set up context defined by expressions $1
    #
    # Usage:
    #     jat_ctx__setup CLASS[.PARAM]=VALUE [CLASS[.PARAM]=VALUE]
    #
    # Set up context consisting of classes and parameters.
    #
    # For every CLASS.PARAM=VALUE expression, a plugin of CLASS
    # must exist, and its setup hadler is called with PARAM=VALUE.
    #
    # See CONTEXT PLUGINS for details on cleanup handlers.
    #
    local expr
    local usage="usage: jat_ctx__setup CLASS[.PARAM]=VALUE [CLASS[.PARAM]=VALUE]"
    local param
    local pname
    local pvalue
    local msg=()
    local __JAT_CTX__FN=jat_ctx__setup
    local __JAT_CTX__ID
    local __JAT_CTX__CLASS
    local re_cls='[[:alpha:]_][[:alnum:]_]*'
    local re_par='@\?[[:alpha:]_][[:alnum:]._]*'
    __JAT_CTX__ID="$(_jat_core__bumpid "context")"
    test $# -gt 0 ||  {
        jat__log_error "$usage"
        return 2
    }
    for expr in "$@"; do
        #shellcheck disable=SC1087
        if grep -qx -e "$re_cls[.]$re_par=.*" <<<"$expr"; then
            IFS='.' read -r __JAT_CTX__CLASS param <<<"$expr"
            IFS='=' read -r pname pvalue <<<"$param"
        elif grep -qx -e "$re_cls=.*" <<<"$expr"; then
            IFS='=' read -r __JAT_CTX__CLASS pvalue <<<"$expr"
            pname=_
        else
            jat__log_error "invalid expression syntax: '$expr'"
            jat__log_error "$usage"
        fi
        _jat_ctx__parama "$pname" "$pvalue"
    done
    _jat_core__rd_keya "contexts/log" "$__JAT_CTX__ID"
    for __JAT_CTX__CLASS in $(__jat_ctx__lscls); do
        jat__log_info "$__JAT_CTX__FN():$__JAT_CTX__ID: setting up: $__JAT_CTX__CLASS"
        _jat_core__rd_keya "contexts/$__JAT_CTX__ID/log" "$__JAT_CTX__CLASS"
        while IFS='=' read -r pname pvalue; do
            test -n "$pname" || continue    #KEEPME: mind the <<<$() newline bug
            jat__log_info "$__JAT_CTX__FN():$__JAT_CTX__ID: .. $pname=$pvalue"
        done <<<"$(__jat_ctx__dmpcparams)"
        while IFS='=' read -r pname pvalue; do
            test -n "$pname" || continue    #KEEPME: mind the <<<$() newline bug
            jat__log_info "$__JAT_CTX__FN():$__JAT_CTX__ID: .. $pname=$pvalue"
            __jat_ctx__call setup "$pname" "$pvalue" || {
                jat__log_error "context setup handler reported error"
                return 3
            }
        done <<<"$(__jat_ctx__dmpparams)"
        if __jat_ctx__has verify; then
            jat__log_info "$__JAT_CTX__FN():$__JAT_CTX__ID: verifying: $__JAT_CTX__CLASS"
            while IFS='=' read -r pname pvalue; do
                test -n "$pname" || continue    #KEEPME: mind the <<<$() newline bug
                __jat_ctx__call verify "$pname" "$pvalue" || {
                    jat__log_error "context verify handler reported error"
                    return 3
                }
            done <<<"$(__jat_ctx__dmpparams)"
        fi
    done
}

_jat_ctx__param() {
    #
    # Print param $1 or _ for context id $__JAT_CTX__ID and class $__JAT_CTX__CLASS
    #
    local want=${1:-_}
    _jat_core__rd_keyR "contexts/$__JAT_CTX__ID/classes/$__JAT_CTX__CLASS/params" \
      | grep -e "^${want//./[.]}=" \
      | cut -d= -f2-
}

_jat_ctx__parama() {
    #
    # Append value $3 to parameter $2 for class $__JAT_CTX__CLASS
    #
    local pname=$1
    local pvalue=$2
    _jat_core__rd_keya "contexts/$__JAT_CTX__ID/classes/$__JAT_CTX__CLASS/params" "$pname=$pvalue"
}

_jat_ctx__think() {
    #
    # think $@ but with useful prefixes
    #
    __jat_ctx__log think "$@"
}

_jat_ctx__thinkc() {
    #
    # think -c $@ but with useful prefixes
    #
    __jat_ctx__logc think "$@"
}

_jat_ctx__thinkf() {
    #
    # think -f $@ but with useful prefixes
    #
    __jat_ctx__logf think "$@"
}

_jat_ctx__thinkv() {
    #
    # think -v $@ but with useful prefixes
    #
    __jat_ctx__logv think "$@"
}

_jat_ctx__warn() {
    #
    # warn $@ but with useful prefixes
    #
    __jat_ctx__log warn "$@"
}

_jat_ctx__warnc() {
    #
    # warn -c $@ but with useful prefixes
    #
    __jat_ctx__logc warn "$@"
}

_jat_ctx__warnf() {
    #
    # warn -f $@ but with useful prefixes
    #
    __jat_ctx__logf warn "$@"
}

_jat_ctx__warnv() {
    #
    # warn -v $@ but with useful prefixes
    #
    __jat_ctx__logv warn "$@"
}


__jat_ctx__call() {
    #
    # Call plugin function $1 for context $__JAT_CTX__ID with args $2..
    #
    local fname=$1; shift
    local dir
    dir=$(_jat_core__rd_path "contexts/$__JAT_CTX__ID/classes/$__JAT_CTX__CLASS/dir")
    (
        #shellcheck disable=SC2030
        export SFPI__PREFIX=_jat_ctxp
        mkdir -p "$dir"
        cd "$dir" || jat__log_error
        sfpi__import "$__JAT_CTX__CLASS" || {
            jat__log_error "failed to import plugin for: $__JAT_CTX__CLASS"
            exit 3
        }
        sfpi__has "$__JAT_CTX__CLASS" "$fname" || {
            jat__log_error "no such plugin function: $fname of $__JAT_CTX__CLASS"
            exit 3
        }
        sfpi__call "$__JAT_CTX__CLASS" "$fname" "$@"
    )
}

__jat_ctx__dmpparams() {
    #
    # Dump all non-empty pre-cached params for context id $__JAT_CTX__ID, class $__JAT_CTX__CLASS
    #
    _jat_core__rd_keyR "contexts/$__JAT_CTX__ID/classes/$__JAT_CTX__CLASS/params" \
      | grep '^[^@]'
}

__jat_ctx__dmpcparams() {
    #
    # Dump all non-empty pre-cached class params for context id $__JAT_CTX__ID, class $__JAT_CTX__CLASS
    #
    local path
    path=$(_jat_core__rd_path "contexts/$__JAT_CTX__ID/classes/$__JAT_CTX__CLASS/params")
    test -e "$path" || return
    _jat_core__rd_keyR "contexts/$__JAT_CTX__ID/classes/$__JAT_CTX__CLASS/params" \
      | grep '^@'
}

__jat_ctx__has() {
    #
    # Check if plugin function $1 is defined by class $__JAT_CTX__CLASS
    #
    local fname=$1; shift
    (
        #shellcheck disable=SC2031
        export SFPI__PREFIX=_jat_ctxp
        sfpi__import "$__JAT_CTX__CLASS" || {
            jat__log_error "failed to import plugin for: $__JAT_CTX__CLASS"
            exit 3
        }
        sfpi__has "$__JAT_CTX__CLASS" "$fname"
    )
}

__jat_ctx__log() {
    #
    # Log lines $2.. using pretty function $1
    #
    local prettyfn=$1; shift
    local lines=("$@")
    local line
    local args=()
    for line in "${lines[@]}"; do
        args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: $line")
    done
    "$prettyfn" -l "${args[@]}"
}

__jat_ctx__log() {
    #
    # Emit decorated log messages $2.. using $1
    #
    local fn=$1; shift
    local msg
    local out=()
    for msg in "$@"; do
        out+=("$__JAT_CTX__FN():$__JAT_CTX__ID: $msg")
    done
    "$fn" "${out[@]}"
}

__jat_ctx__logc() {
    #
    # Log output of command $2.. using pretty function $1
    #
    local prettyfn=$1; shift
    local cmd=("$@")
    local line
    local args=()
    local cout
    local cerr
    local ces
    cout=$(mktemp -t __jat_ctx__logc.cout.XXXXXXXX)
    cerr=$(mktemp -t __jat_ctx__logc.cerr.XXXXXXXX)
    args=("$__JAT_CTX__FN():$__JAT_CTX__ID: --- begin command: ${cmd[*]} ---")
    "${cmd[@]}" >"$cout" 2>"$cerr"; ces=$?
    while read -r line; do
        args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: stderr|$line")
    done <"$cerr"
    while read -r line; do
        args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: stdout|$line")
    done <"$cout"
    args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: --- end command; es=$ces ---")
    "$prettyfn" -l "${args[@]}"
    rm "$cout" "$cerr"
    return "$ces"
}

__jat_ctx__logv() {
    #
    # Log values of variables $2.. using pretty function $1
    #
    local prettyfn=$1; shift
    local vars=("$@")
    local var
    local decl
    local args=()
    for var in "${vars[@]}"; do
        if decl=$(declare -p "$var" 2>/dev/null); then
            args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: ${decl:11}")
        else
            args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: $var #undefined")
        fi
    done
    "$prettyfn" -l "${args[@]}"
}

__jat_ctx__logf() {
    #
    # Log content of files $2.. using pretty function $1
    #
    local prettyfn=$1; shift
    local paths=("$@")
    local path
    local line
    local args=()
    for path in "${paths[@]}"; do
        test -e "$path" || {
            args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: --- nonexistent file: $path ---")
            continue
        }
        test -s "$path" || {
            args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: --- empty file: $path ---")
            continue
        }
        args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: --- begin file: $path ---")
        while read -r line; do
            args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: |$line")
        done <"$path"
        args+=("$__JAT_CTX__FN():$__JAT_CTX__ID: --- end file: $path ---")
    done
    "$prettyfn" -l "${args[@]}"
}

__jat_ctx__lscls() {
    #
    # List all pre-cached classes for context id $__JAT_CTX__ID
    #
    local path
    path=$(_jat_core__rd_path "contexts/$__JAT_CTX__ID/classes")
    find "$path" -type d \
        -mindepth 1 -maxdepth 1 \
        -printf '%P\n' 2>/dev/null \
      | LC_ALL=C sort
}

#shellfu module-version=0.2.12
